import math
import os
import sys

from .hmmlearn import HMMLearn

dir_path = os.path.dirname(os.path.realpath(__file__))
parent_dir_path = os.path.abspath(os.path.join(dir_path, os.pardir))
sys.path.insert(0, parent_dir_path)


class HMMDecode:

    def __init__(self, hmm, log=False):
        self.hmm = hmm
        self.log = log

    def decode(self, sentence, tags=None):
        if tags is not None:
            tags = [s.split("/") for s in tags]
        else:
            tags = [self.hmm.TAGS for _ in range(len(sentence))]

        if self.hmm.trigram:
            return self.__decode_trigram(sentence, tags)
        else:
            return self.__decode_bigram(sentence, tags)

    def __decode_bigram(self, sentence, tags):
        viterbi_tag = {}
        viterbi_backpointer = {}
        if len(tags[0]) == 1:
            tag = tags[0][0]
            if self.log:
                viterbi_tag[tag] = 0.0
            else:
                viterbi_tag[tag] = 1.0
            viterbi_backpointer[tag] = "START"
        else:
            for tag in tags[0]:
                if self.log:
                    viterbi_tag[tag] = math.log(self.hmm.get_transition_prob(tag, "START")) + \
                                       math.log(self.hmm.get_emission_prob(sentence[0], tag))
                else:
                    viterbi_tag[tag] = self.hmm.get_transition_prob(tag, "START") * \
                                       self.hmm.get_emission_prob(sentence[0], tag)
                viterbi_backpointer[tag] = "START"

        viterbi_main = []
        backpointer_main = []

        viterbi_main.append(viterbi_tag)
        backpointer_main.append(viterbi_backpointer)
        return self.__viterbi(sentence, tags, viterbi_main, backpointer_main)

    def __decode_trigram(self, sentence, tags):
        viterbi_tag = {}
        viterbi_backpointer = {}
        if len(tags[0]) == 1:
            tag = tags[0][0]
            if self.log:
                viterbi_tag[tag] = 0.0
            else:
                viterbi_tag[tag] = 1.0
            viterbi_backpointer[tag] = ("END", "START")
        else:
            for tag in tags[0]:
                if self.log:
                    viterbi_tag[tag] = math.log(self.hmm.get_transition_prob(tag, ("END", "START"))) + \
                                       math.log(self.hmm.get_emission_prob(sentence[0], tag))
                else:
                    viterbi_tag[tag] = self.hmm.get_transition_prob(tag, ("END", "START")) * \
                                       self.hmm.get_emission_prob(sentence[0], tag)
                viterbi_backpointer[tag] = ("END", "START")

        viterbi = []
        backpointer = []

        viterbi.append(viterbi_tag)
        backpointer.append(viterbi_backpointer)
        return self.__viterbi(sentence, tags, viterbi, backpointer)

    def __viterbi(self, sentence, tags, viterbi, backpointer):
        N = len(sentence)
        for index in range(1, N):
            cur_viterbi = {}
            cur_backpointer = {}
            prev_viterbi = viterbi[-1]

            if len(tags[index]) == 1:
                prev_best = max(prev_viterbi.keys(), key=lambda prevtag: prev_viterbi[prevtag])
                tag = tags[index][0]
                cur_viterbi[tag] = prev_viterbi[prev_best]
                cur_backpointer[tag] = prev_best
            else:
                for tag in tags[index]:
                    prev_best = max(prev_viterbi.keys(),
                                    key=lambda prev_tag:
                                        self.__calculate_viterbi_value(
                                            prev_viterbi[prev_tag],
                                            self.hmm.get_transition_prob(tag, prev_tag),
                                            self.hmm.get_emission_prob(sentence[index], tag)))

                    cur_viterbi[tag] = self.__calculate_viterbi_value(
                        prev_viterbi[prev_best],
                        self.hmm.get_transition_prob(tag, prev_best),
                        self.hmm.get_emission_prob(sentence[index], tag))
                    cur_backpointer[tag] = prev_best

            viterbi.append(cur_viterbi)
            backpointer.append(cur_backpointer)

        prev_viterbi = viterbi[-1]
        prev_best = max(prev_viterbi.keys(),
                        key=lambda prev_tag: self.__calculate_viterbi_value(
                            prev_viterbi[prev_tag],
                            self.hmm.get_transition_prob("END", prev_tag),
                            1.0
                        ))

        # print("viterbi:")
        # print(viterbi)
        # print("backpointer:")
        # print(backpointer)
        # prob_tag_sequence = prev_viterbi[prev_best] * self.hmm.get_transition_prob("END", prev_best)
        # print("prob_tag_sequence: ", prob_tag_sequence)

        best_tag_sequence = ["END", prev_best]
        backpointer.reverse()

        current_best_tag = prev_best
        for pointer in backpointer:
            best_tag_sequence.append(pointer[current_best_tag])
            current_best_tag = pointer[current_best_tag]

        best_tag_sequence.reverse()
        return best_tag_sequence[1:-1]

    def __calculate_viterbi_value(self, prev_prob, trans_prob, emission_prob):
        if self.log:
            return prev_prob + math.log(trans_prob) + math.log(emission_prob)
        else:
            return prev_prob * trans_prob * emission_prob


if __name__ == "__main__":
    hmmlearn = HMMLearn(trigram=True)
    sample_sentence = ["Mereka", "membaca", "banyak", "sekali", "buku", "."]
    sample_tags = ["VERB/PRON", "VERB", "DET/ADJ", "DET/ADV", "NOUN", "PUNCT"]
    hmmdecode = HMMDecode(hmmlearn, log=True)
    best_tag = hmmdecode.decode(sample_sentence)
    print(best_tag)



import os
from statistics import mode
import gdown
import zipfile


class ModelProxy:
    file_id = {
        "FR_GSD-ID_CSUI": "1tg3qNum3Q97-ogBJbXeSqwEdV38kN1hu",
        "IT_ISDT-ID_CSUI": "1K4-mwq9CwHjoIHfGksUUVghP35oyUB8b",
        "SL_SSJ-ID_CSUI": "1jzIJMUWbuueMOWHklQthQAzHVKw4tacT",
        "EN_GUM-ID_CSUI": "1PmzgzT4F5_gBUtMPE8ZEJNGJOTjkc9tj",
        "FR_GSD-ID_GSD": "1bpJsiPwRI09PwA8FdH1L7n1BPw4WbLGw",
        "IT_ISDT-ID_GSD": "1n8VE-LiKkdsUvyNvREK0IEVoNIZ0mn5b",
        "SL_SSJ-ID_GSD": "1IoAYZdKkKCHR4-azXQsg2gQrAX-vtPuR",
        "EN_GUM-ID_GSD": "1JXwgsMwkhfRtvNx9q6hGH-izyA8f-mph",
    }

    def download_model(model_name):
        if model_name not in ModelProxy.file_id:
            print(model_name)
            raise NotImplementedError

        url = "https://drive.google.com/uc?id={}".format(ModelProxy.file_id[model_name])
        curdir_path, _ = os.path.split(os.path.realpath(__file__))
        output_dir = os.path.join(curdir_path, ".pretrained_model")
        output_file = os.path.join(output_dir, "{}.zip".format(model_name))
        gdown.download(url, output_file, quiet=True)

        with zipfile.ZipFile(output_file, "r") as zip_ref:
            zip_ref.extractall(output_dir)



from typing import List, Literal
import re
import os

from ..conllu import ConlluData

def read_conllu(file_path: str, separator: str=r'\s+') -> List[List[ConlluData]]:
    conllu_result = []
    with open(file_path, encoding='utf-8') as conllu_file:
        next_sentence = conllu_file.readline()
        while next_sentence != '':

            # next_line currently is '# sent_id = ...'
            conllu_file.readline()  #  '# read text = ...'

            one_sentence_result = []
            next_line = conllu_file.readline()
            while next_line not in ['', '\n']:
                conllu_attrs= re.split(separator, next_line)
                one_sentence_result.append(ConlluData(*conllu_attrs[:8]))
                next_line = conllu_file.readline()

            conllu_result.append(one_sentence_result)
            next_sentence = conllu_file.readline()

    return conllu_result


def write_conllu(list_list_conllu: List[List[ConlluData]], file_path: str,
                 write_mode: Literal['a', 'w', 'x'] = 'x', separator: str='\t'
                 ) -> str:

    all_write_modes = ['a', 'w', 'x']

    if write_mode not in all_write_modes:
        raise ValueError(f"write_mode must be one of {all_write_modes}, but {write_mode} was given")

    with open(file_path, mode=write_mode, encoding='utf-8') as file:
        if write_mode == 'a':
            file.writelines('\n')

        for sentence_idx, list_conllu in enumerate(list_list_conllu):
            file.writelines(f'# sent_id = {sentence_idx + 1}\n')

            sentence = " ".join(map(lambda x: x.get_form(), list_conllu))
            file.writelines(f'# text = {sentence}\n')

            for row_idx, conllu in enumerate(list_conllu):
                _, *row = str(conllu).split('\t')
                row_sentence = separator.join(row)
                file.writelines(f"{row_idx + 1}{separator}{row_sentence}\n")

            if sentence_idx < len(list_list_conllu) - 1:
                file.writelines('\n')

    return os.path.realpath(file_path)



#!/usr/bin/python3

from disambiguation.hmmlearn import HMMLearn
from disambiguation.hmmdecode import HMMDecode


class Disambiguator:

    def __init__(self):
        self.__hmmlearn = HMMLearn(trigram=True)
        self.__hmmdecode = HMMDecode(hmm=self.__hmmlearn, log=True)

    def disambiguate(self, rows):
        sentences = []
        tags = []
        for row in rows:
            sentences.append(row[1])
            tags.append(row[3])

        predicted_tags = self.__hmmdecode.decode(sentences, tags)
        new_rows = []
        for i in range(len(rows)):
            row = rows[i]
            temp = row.copy()
            lemmas = temp[2].split("/")
            tags = temp[3].split("/")
            features = temp[5].split("/")
            misc = temp[9].split("/")
            if len(tags) > 1:
                for j in range(len(tags)):
                    if tags[j] == predicted_tags[i]:
                        temp[2] = lemmas[j]
                        temp[3] = tags[j]
                        temp[9] = misc[j]
                        for feature in features:
                            temp[5] = "_"
                            if tags[j] in feature:
                                idx = feature.find("->")
                                temp[5] = feature[idx+3:-1]
                                break

                        break

            new_rows.append(temp)

        return new_rows



"""
Implementation of "Attention is All You Need"
"""

import torch.nn as nn

from .util_class import LayerNorm
from .multi_head_attn import MultiHeadedAttention
from .encoder import EncoderBase
from .position_ffn import PositionwiseFeedForward


class TransformerEncoderLayer(nn.Module):
    """
    A single layer of the transformer encoder.
    Args:
        d_model (int): the dimension of keys/values/queries in
                   MultiHeadedAttention, also the input size of
                   the first-layer of the PositionwiseFeedForward.
        heads (int): the number of head for MultiHeadedAttention.
        d_ff (int): the second-layer of the PositionwiseFeedForward.
        dropout (float): dropout probability(0-1.0).
    """

    def __init__(self, d_model, heads, d_ff, d_k, d_v,
                 attn_drop, relu_drop, res_drop, clip_dist=0, use_neg_dist=False):
        super(TransformerEncoderLayer, self).__init__()

        self.self_attn = MultiHeadedAttention(heads, d_model, d_k, d_v,
                                              dropout=attn_drop,
                                              clip_dist=clip_dist, use_neg_dist=use_neg_dist)
        self.feed_forward = PositionwiseFeedForward(d_model, d_ff, relu_drop, res_drop)
        self.layer_norm = LayerNorm(d_model)
        self.dropout = nn.Dropout(res_drop)

    def forward(self, inputs, mask):
        """
        Transformer Encoder Layer definition.
        Args:
            inputs (`FloatTensor`): `[batch_size x src_len x model_dim]`
            mask (`LongTensor`): `[batch_size x src_len x src_len]`
        Returns:
            (`FloatTensor`):
            * outputs `[batch_size x src_len x model_dim]`
        """
        input_norm = self.layer_norm(inputs)
        context, _ = self.self_attn(input_norm, input_norm, input_norm, mask=mask)
        out = self.dropout(context) + inputs
        return self.feed_forward(out)


class TransformerEncoder(EncoderBase):
    """
    The Transformer encoder from "Attention is All You Need".
    .. mermaid::
       graph BT
          A[input]
          B[multi-head self-attn]
          C[feed forward]
          O[output]
          A --> B
          B --> C
          C --> O
    Args:
        num_layers (int): number of encoder layers
        d_model (int): size of the model
        heads (int): number of heads
        d_ff (int): size of the inner FF layer
        dropout (float): dropout parameters
        embeddings (:obj:`onmt.modules.Embeddings`):
          embeddings to use, should have positional encodings
    Returns:
        (`FloatTensor`, `FloatTensor`):
        * embeddings `[batch_size x src_len x model_dim]`
        * memory_bank `[batch_size x src_len x model_dim]`
    """

    def __init__(self, num_layers, d_model, heads,
                 d_ff, d_k, d_v,
                 attn_drop, relu_drop, res_drop,
                 clip_dist=0, use_neg_dist=False):
        super(TransformerEncoder, self).__init__()

        self.num_layers = num_layers
        self.transformer = nn.ModuleList(
            [TransformerEncoderLayer(d_model, heads, d_ff, d_k, d_v,
                                     attn_drop, relu_drop, res_drop,
                                     clip_dist=clip_dist, use_neg_dist=use_neg_dist)
             for _ in range(num_layers)])
        self.layer_norm = LayerNorm(d_model)

    def forward(self, emb, lengths=None):
        """ See :obj:`EncoderBase.forward()`"""
        self._check_args(emb, lengths)

        out = emb
        # Run the forward pass of every layer of the tranformer.
        for i in range(self.num_layers):
            out = self.transformer[i](out, mask=None)
        out = self.layer_norm(out)

        return out



__author__ = 'max'

import numpy as np
from enum import Enum
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
from ..nn import VarMaskedFastLSTM
from ..nn import BiAAttention, BiLinear
from ..tasks import parser
from ..transformer import TransformerEncoder

class PriorOrder(Enum):
    DEPTH = 0
    INSIDE_OUT = 1
    LEFT2RIGTH = 2


class BiRecurrentConvBiAffine(nn.Module):
    def __init__(self, word_dim, num_words, char_dim, num_chars, pos_dim, num_pos, num_filters, kernel_size, rnn_mode,
                 hidden_size, num_layers, num_labels, arc_space, type_space,
                 embedd_word=None, embedd_char=None, embedd_pos=None, p_in=0.33, p_out=0.33, p_rnn=(0.33, 0.33),
                 biaffine=True, pos=True, char=True,
                 train_position=False, use_con_rnn=True, trans_hid_size=1028, d_k=64, d_v=64, multi_head_attn=True,
                 num_head=8, enc_use_neg_dist=False, enc_clip_dist=0, position_dim=50, max_sent_length=200,
                 use_gpu=False, no_word=False,
                 ):
        super(BiRecurrentConvBiAffine, self).__init__()

        self.word_embedd = nn.Embedding(num_words, word_dim, _weight=embedd_word)
        self.pos_embedd = nn.Embedding(num_pos, pos_dim, _weight=embedd_pos) if pos else None
        self.char_embedd = nn.Embedding(num_chars, char_dim, _weight=embedd_char) if char else None

        self.conv1d = nn.Conv1d(char_dim, num_filters, kernel_size, padding=kernel_size - 1) if char else None
        self.dropout_in = nn.Dropout2d(p=p_in)
        self.dropout_out = nn.Dropout2d(p=p_out)
        self.num_labels = num_labels
        self.pos = pos
        self.char = char
        self.no_word = no_word
        #
        self.use_con_rnn = use_con_rnn
        self.multi_head_attn = multi_head_attn
        self.use_gpu = use_gpu
        self.position_dim = position_dim

        if rnn_mode == 'FastLSTM':
            RNN = VarMaskedFastLSTM

        else:
            raise ValueError('Unknown RNN mode: %s' % rnn_mode)

        dim_enc = 0
        if not no_word:
            dim_enc = word_dim
        if pos:
            dim_enc += pos_dim
        if char:
            dim_enc += num_filters

        #
        self.encoder_layers = num_layers
        if self.use_con_rnn:
            self.rnn = RNN(dim_enc, hidden_size, num_layers=self.encoder_layers, batch_first=True,
                                       bidirectional=True, dropout=p_rnn)
            enc_output_dim = 2 * hidden_size
        else:
            if self.multi_head_attn:
                pos_emb_size = position_dim
                d_model = pos_emb_size + dim_enc
                if position_dim > 0:
                    self.position_embedding = nn.Embedding(max_sent_length, pos_emb_size)
                    if not train_position:
                        self.position_embedding.weight.requires_grad = False  # turn off pos embedding training
                        ######################### init positional embedding ##########################
                        # keep dim 0 for padding token position encoding zero vector
                        position_enc = np.array([[pos / np.power(10000, 2 * (j // 2) / pos_emb_size) for j in
                                                  range(pos_emb_size)] if pos != 0 else np.zeros(pos_emb_size) for pos in
                                                 range(max_sent_length)])
                        position_enc[1:, 0::2] = np.sin(position_enc[1:, 0::2])  # dim 2i
                        position_enc[1:, 1::2] = np.cos(position_enc[1:, 1::2])  # dim 2i+1
                        self.position_embedding.weight.data.copy_(torch.from_numpy(position_enc).type(torch.FloatTensor))
                        ##############################################################################
                #
                self.transformer = TransformerEncoder(self.encoder_layers,
                                                      d_model=d_model,
                                                      heads=num_head,
                                                      d_ff=trans_hid_size,
                                                      d_k=d_k,
                                                      d_v=d_v,
                                                      attn_drop=p_rnn[0],
                                                      relu_drop=p_rnn[1],
                                                      res_drop=p_rnn[2],
                                                      clip_dist=enc_clip_dist,
                                                      use_neg_dist=enc_use_neg_dist)

                enc_output_dim = d_model
            else:
                raise NotImplementedError()

        # self.rnn = RNN(dim_enc, hidden_size, num_layers=num_layers, batch_first=True, bidirectional=True, dropout=p_rnn)

        out_dim = enc_output_dim
        self.arc_h = nn.Linear(out_dim, arc_space)
        self.arc_c = nn.Linear(out_dim, arc_space)
        self.attention = BiAAttention(arc_space, arc_space, 1, biaffine=biaffine)

        self.type_h = nn.Linear(out_dim, type_space)
        self.type_c = nn.Linear(out_dim, type_space)
        self.bilinear = BiLinear(type_space, type_space, self.num_labels)

    def _get_rnn_output(self, input_word, input_char, input_pos, mask=None, length=None, hx=None):
        input = None

        if not self.no_word:
            # [batch, length, word_dim]
            word = self.word_embedd(input_word)
            # apply dropout on input
            word = self.dropout_in(word)

            input = word

        if self.char:
            # [batch, length, char_length, char_dim]
            char = self.char_embedd(input_char)
            char_size = char.size()
            # first transform to [batch *length, char_length, char_dim]
            # then transpose to [batch * length, char_dim, char_length]
            char = char.view(char_size[0] * char_size[1], char_size[2], char_size[3]).transpose(1, 2)
            # put into cnn [batch*length, char_filters, char_length]
            # then put into maxpooling [batch * length, char_filters]
            char, _ = self.conv1d(char).max(dim=2)
            # reshape to [batch, length, char_filters]
            char = torch.tanh(char).view(char_size[0], char_size[1], -1)
            # apply dropout on input
            char = self.dropout_in(char)
            # concatenate word and char [batch, length, word_dim+char_filter]
            input = char if input is None else torch.cat([input, char], dim=2)

        if self.pos:
            # [batch, length, pos_dim]
            pos = self.pos_embedd(input_pos)
            # # apply dropout on input
            # pos = self.dropout_in(pos)
            input = pos if input is None else torch.cat([input, pos], dim=2)

        # # output from rnn [batch, length, hidden_size]
        # output, hn = self.rnn(input, mask, hx=hx)

        if self.use_con_rnn:
            output, hn = self.rnn(input, mask, hx=hx)
        else:
            if self.multi_head_attn:
                src_encoding = input
                if self.position_dim > 0:
                    position_encoding = Variable(torch.arange(start=0, end=src_encoding.size(1)).type(torch.LongTensor))
                    # ----- modified by zs
                    if self.use_gpu:
                        position_encoding = position_encoding.cuda()
                    # -----
                    position_encoding = position_encoding.expand(*src_encoding.size()[:-1])
                    position_encoding = self.position_embedding(position_encoding)
                    # src_encoding = src_encoding + position_encoding
                    src_encoding = torch.cat([src_encoding, position_encoding], dim=2)
                src_encoding = self.transformer(src_encoding)
                output, hn = src_encoding, None
            else:
                raise NotImplementedError()

        # apply dropout for output
        # [batch, length, hidden_size] --> [batch, hidden_size, length] --> [batch, length, hidden_size]
        output = self.dropout_out(output.transpose(1, 2)).transpose(1, 2)

        # output size [batch, length, arc_space]
        arc_h = F.elu(self.arc_h(output))
        arc_c = F.elu(self.arc_c(output))

        # output size [batch, length, type_space]
        type_h = F.elu(self.type_h(output))
        type_c = F.elu(self.type_c(output))

        # apply dropout
        # [batch, length, dim] --> [batch, 2 * length, dim]
        arc = torch.cat([arc_h, arc_c], dim=1)
        type = torch.cat([type_h, type_c], dim=1)

        arc = self.dropout_out(arc.transpose(1, 2)).transpose(1, 2)
        arc_h, arc_c = arc.chunk(2, 1)

        type = self.dropout_out(type.transpose(1, 2)).transpose(1, 2)
        type_h, type_c = type.chunk(2, 1)
        type_h = type_h.contiguous()
        type_c = type_c.contiguous()

        return (arc_h, arc_c), (type_h, type_c), hn, mask, length

    def forward(self, input_word, input_char, input_pos, mask=None, length=None, hx=None):
        # output from rnn [batch, length, tag_space]
        arc, type, _, mask, length = self._get_rnn_output(input_word, input_char, input_pos, mask=mask, length=length,
                                                          hx=hx)
        # [batch, length, length]
        out_arc = self.attention(arc[0], arc[1], mask_d=mask, mask_e=mask).squeeze(dim=1)
        return out_arc, type, mask, length

    def loss(self, input_word, input_char, input_pos, heads, types, mask=None, length=None, hx=None):
        # out_arc shape [batch, length, length]
        out_arc, out_type, mask, length = self.forward(input_word, input_char, input_pos, mask=mask, length=length,
                                                       hx=hx)
        batch, max_len, _ = out_arc.size()

        if length is not None and heads.size(1) != mask.size(1):
            heads = heads[:, :max_len]
            types = types[:, :max_len]

        # out_type shape [batch, length, type_space]
        type_h, type_c = out_type

        # create batch index [batch]
        batch_index = torch.arange(0, batch).type_as(out_arc.data).long()
        # get vector for heads [batch, length, type_space],
        type_h = type_h[batch_index, heads.data.t()].transpose(0, 1).contiguous()
        # compute output for type [batch, length, num_labels]
        out_type = self.bilinear(type_h, type_c)

        # mask invalid position to -inf for log_softmax
        if mask is not None:
            minus_inf = -1e8
            minus_mask = (1 - mask) * minus_inf
            out_arc = out_arc + minus_mask.unsqueeze(2) + minus_mask.unsqueeze(1)

        # loss_arc shape [batch, length, length]
        loss_arc = F.log_softmax(out_arc, dim=1)
        # loss_type shape [batch, length, num_labels]
        loss_type = F.log_softmax(out_type, dim=2)

        # mask invalid position to 0 for sum loss
        if mask is not None:
            loss_arc = loss_arc * mask.unsqueeze(2) * mask.unsqueeze(1)
            loss_type = loss_type * mask.unsqueeze(2)
            # number of valid positions which contribute to loss (remove the symbolic head for each sentence.
            num = mask.sum() - batch
        else:
            # number of valid positions which contribute to loss (remove the symbolic head for each sentence.
            num = float(max_len - 1) * batch

        # first create index matrix [length, batch]
        child_index = torch.arange(0, max_len).view(max_len, 1).expand(max_len, batch)
        child_index = child_index.type_as(out_arc.data).long()
        # [length-1, batch]
        loss_arc = loss_arc[batch_index, heads.data.t(), child_index][1:]
        loss_type = loss_type[batch_index, child_index, types.data.t()][1:]

        return -loss_arc.sum() / num, -loss_type.sum() / num

    def _decode_types(self, out_type, heads, leading_symbolic):
        # out_type shape [batch, length, type_space]
        type_h, type_c = out_type
        batch, max_len, _ = type_h.size()
        # create batch index [batch]
        batch_index = torch.arange(0, batch).type_as(type_h.data).long()
        # get vector for heads [batch, length, type_space],
        type_h = type_h[batch_index, heads.t()].transpose(0, 1).contiguous()
        # compute output for type [batch, length, num_labels]
        out_type = self.bilinear(type_h, type_c)
        # remove the first #leading_symbolic types.
        out_type = out_type[:, :, leading_symbolic:]
        # compute the prediction of types [batch, length]
        _, types = out_type.max(dim=2)
        return types + leading_symbolic

    def decode(self, input_word, input_char, input_pos, mask=None, length=None, hx=None, leading_symbolic=0):
        # out_arc shape [batch, length, length]
        out_arc, out_type, mask, length = self.forward(input_word, input_char, input_pos, mask=mask, length=length,
                                                       hx=hx)
        out_arc = out_arc.data
        batch, max_len, _ = out_arc.size()
        # set diagonal elements to -inf
        out_arc = out_arc + torch.diag(out_arc.new(max_len).fill_(-np.inf))
        # set invalid positions to -inf
        if mask is not None:
            # minus_mask = (1 - mask.data).byte().view(batch, max_len, 1)
            minus_mask = (1 - mask.data).byte().unsqueeze(2)
            out_arc.masked_fill_(minus_mask, -np.inf)

        # compute naive predictions.
        # predition shape = [batch, length]
        _, heads = out_arc.max(dim=1)

        types = self._decode_types(out_type, heads, leading_symbolic)

        return heads.cpu().numpy(), types.data.cpu().numpy()

    def decode_mst(self, input_word, input_char, input_pos, mask=None, length=None, hx=None, leading_symbolic=0):
        '''
        Args:
            input_word: Tensor
                the word input tensor with shape = [batch, length]
            input_char: Tensor
                the character input tensor with shape = [batch, length, char_length]
            input_pos: Tensor
                the pos input tensor with shape = [batch, length]
            mask: Tensor or None
                the mask tensor with shape = [batch, length]
            length: Tensor or None
                the length tensor with shape = [batch]
            hx: Tensor or None
                the initial states of RNN
            leading_symbolic: int
                number of symbolic labels leading in type alphabets (set it to 0 if you are not sure)

        Returns: (Tensor, Tensor)
                predicted heads and types.

        '''
        # out_arc shape [batch, length, length]
        out_arc, out_type, mask, length = self.forward(input_word, input_char, input_pos, mask=mask, length=length,
                                                       hx=hx)

        # out_type shape [batch, length, type_space]
        type_h, type_c = out_type
        batch, max_len, type_space = type_h.size()

        # compute lengths
        if length is None:
            if mask is None:
                length = [max_len for _ in range(batch)]
            else:
                length = mask.data.sum(dim=1).long().cpu().numpy()

        type_h = type_h.unsqueeze(2).expand(batch, max_len, max_len, type_space).contiguous()
        type_c = type_c.unsqueeze(1).expand(batch, max_len, max_len, type_space).contiguous()
        # compute output for type [batch, length, length, num_labels]
        out_type = self.bilinear(type_h, type_c)

        # mask invalid position to -inf for log_softmax
        if mask is not None:
            minus_inf = -1e8
            minus_mask = (1 - mask) * minus_inf
            out_arc = out_arc + minus_mask.unsqueeze(2) + minus_mask.unsqueeze(1)

        # loss_arc shape [batch, length, length]
        loss_arc = F.log_softmax(out_arc, dim=1)
        # loss_type shape [batch, length, length, num_labels]
        loss_type = F.log_softmax(out_type, dim=3).permute(0, 3, 1, 2)
        # [batch, num_labels, length, length]
        energy = torch.exp(loss_arc.unsqueeze(1) + loss_type)

        return parser.decode_MST(energy.data.cpu().numpy(), length, leading_symbolic=leading_symbolic, labeled=True)



__author__ = 'max'

import re
import numpy as np

def is_uni_punctuation(word):
    match = re.match("^[^\w\s]+$]", word, flags=re.UNICODE)
    return match is not None


def is_punctuation(word, pos, punct_set=None):
    if punct_set is None:
        return is_uni_punctuation(word)
    else:
        return pos in punct_set


def eval(words, postags, heads_pred, types_pred, heads, types, word_alphabet, pos_alphabet, lengths,
         punct_set=None, symbolic_root=False, symbolic_end=False):
    batch_size, _ = words.shape
    ucorr = 0.
    lcorr = 0.
    total = 0.
    ucomplete_match = 0.
    lcomplete_match = 0.

    ucorr_nopunc = 0.
    lcorr_nopunc = 0.
    total_nopunc = 0.
    ucomplete_match_nopunc = 0.
    lcomplete_match_nopunc = 0.

    corr_root = 0.
    total_root = 0.
    start = 1 if symbolic_root else 0
    end = 1 if symbolic_end else 0
    for i in range(batch_size):
        ucm = 1.
        lcm = 1.
        ucm_nopunc = 1.
        lcm_nopunc = 1.

        print(i)
        for j in range(start, lengths[i] - end):
            word = word_alphabet.get_instance(words[i, j])
            word = word.encode('utf8')

            pos = pos_alphabet.get_instance(postags[i, j])
            pos = pos.encode('utf8')

            total += 1
            if heads[i, j] == heads_pred[i, j]:
                ucorr += 1
                if types[i, j] == types_pred[i, j]:
                    lcorr += 1
                else:
                    lcm = 0
            else:
                ucm = 0
                lcm = 0

            if not is_punctuation(word, pos, punct_set):
                total_nopunc += 1
                if heads[i, j] == heads_pred[i, j]:
                    ucorr_nopunc += 1
                    if types[i, j] == types_pred[i, j]:
                        lcorr_nopunc += 1
                    else:
                        lcm_nopunc = 0
                else:
                    ucm_nopunc = 0
                    lcm_nopunc = 0

            if heads[i, j] == 0:
                total_root += 1
                corr_root += 1 if heads_pred[i, j] == 0 else 0

        ucomplete_match += ucm
        lcomplete_match += lcm
        ucomplete_match_nopunc += ucm_nopunc
        lcomplete_match_nopunc += lcm_nopunc

    return (ucorr, lcorr, total, ucomplete_match, lcomplete_match), \
           (ucorr_nopunc, lcorr_nopunc, total_nopunc, ucomplete_match_nopunc, lcomplete_match_nopunc), \
           (corr_root, total_root), batch_size


def decode_MST(energies, lengths, leading_symbolic=0, labeled=True):
    """
    decode best parsing tree with MST algorithm.
    :param energies: energies: numpy 4D tensor
        energies of each edge. the shape is [batch_size, num_labels, n_steps, n_steps],
        where the summy root is at index 0.
    :param masks: numpy 2D tensor
        masks in the shape [batch_size, n_steps].
    :param leading_symbolic: int
        number of symbolic dependency types leading in type alphabets)
    :return:
    """

    def find_cycle(par):
        added = np.zeros([length], bool)
        added[0] = True
        cycle = set()
        findcycle = False
        for i in range(1, length):
            if findcycle:
                break

            if added[i] or not curr_nodes[i]:
                continue

            # init cycle
            tmp_cycle = set()
            tmp_cycle.add(i)
            added[i] = True
            findcycle = True
            l = i

            while par[l] not in tmp_cycle:
                l = par[l]
                if added[l]:
                    findcycle = False
                    break
                added[l] = True
                tmp_cycle.add(l)

            if findcycle:
                lorg = l
                cycle.add(lorg)
                l = par[lorg]
                while l != lorg:
                    cycle.add(l)
                    l = par[l]
                break

        return findcycle, cycle

    def chuLiuEdmonds():
        par = np.zeros([length], dtype=np.int32)
        # create best graph
        par[0] = -1
        for i in range(1, length):
            # only interested at current nodes
            if curr_nodes[i]:
                max_score = score_matrix[0, i]
                par[i] = 0
                for j in range(1, length):
                    if j == i or not curr_nodes[j]:
                        continue

                    new_score = score_matrix[j, i]
                    if new_score > max_score:
                        max_score = new_score
                        par[i] = j

        # find a cycle
        findcycle, cycle = find_cycle(par)
        # no cycles, get all edges and return them.
        if not findcycle:
            final_edges[0] = -1
            for i in range(1, length):
                if not curr_nodes[i]:
                    continue

                pr = oldI[par[i], i]
                ch = oldO[par[i], i]
                final_edges[ch] = pr
            return

        cyc_len = len(cycle)
        cyc_weight = 0.0
        cyc_nodes = np.zeros([cyc_len], dtype=np.int32)
        id = 0
        for cyc_node in cycle:
            cyc_nodes[id] = cyc_node
            id += 1
            cyc_weight += score_matrix[par[cyc_node], cyc_node]

        rep = cyc_nodes[0]
        for i in range(length):
            if not curr_nodes[i] or i in cycle:
                continue

            max1 = float("-inf")
            wh1 = -1
            max2 = float("-inf")
            wh2 = -1

            for j in range(cyc_len):
                j1 = cyc_nodes[j]
                if score_matrix[j1, i] > max1:
                    max1 = score_matrix[j1, i]
                    wh1 = j1

                scr = cyc_weight + score_matrix[i, j1] - score_matrix[par[j1], j1]

                if scr > max2:
                    max2 = scr
                    wh2 = j1

            score_matrix[rep, i] = max1
            oldI[rep, i] = oldI[wh1, i]
            oldO[rep, i] = oldO[wh1, i]
            score_matrix[i, rep] = max2
            oldO[i, rep] = oldO[i, wh2]
            oldI[i, rep] = oldI[i, wh2]

        rep_cons = []
        for i in range(cyc_len):
            rep_cons.append(set())
            cyc_node = cyc_nodes[i]
            for cc in reps[cyc_node]:
                rep_cons[i].add(cc)

        for i in range(1, cyc_len):
            cyc_node = cyc_nodes[i]
            curr_nodes[cyc_node] = False
            for cc in reps[cyc_node]:
                reps[rep].add(cc)

        chuLiuEdmonds()

        # check each node in cycle, if one of its representatives is a key in the final_edges, it is the one.
        found = False
        wh = -1
        for i in range(cyc_len):
            for repc in rep_cons[i]:
                if repc in final_edges:
                    wh = cyc_nodes[i]
                    found = True
                    break
            if found:
                break

        l = par[wh]
        while l != wh:
            ch = oldO[par[l], l]
            pr = oldI[par[l], l]
            final_edges[ch] = pr
            l = par[l]

    if labeled:
        assert energies.ndim == 4, 'dimension of energies is not equal to 4'
    else:
        assert energies.ndim == 3, 'dimension of energies is not equal to 3'
    input_shape = energies.shape
    batch_size = input_shape[0]
    max_length = input_shape[2]

    pars = np.zeros([batch_size, max_length], dtype=np.int32)
    types = np.zeros([batch_size, max_length], dtype=np.int32) if labeled else None
    for i in range(batch_size):
        energy = energies[i]

        # calc the realy length of this instance
        length = lengths[i]

        # calc real energy matrix shape = [length, length, num_labels - #symbolic] (remove the label for symbolic types).
        if labeled:
            energy = energy[leading_symbolic:, :length, :length]
            # get best label for each edge.
            label_id_matrix = energy.argmax(axis=0) + leading_symbolic
            energy = energy.max(axis=0)
        else:
            energy = energy[:length, :length]
            label_id_matrix = None
        # get original score matrix
        orig_score_matrix = energy
        # initialize score matrix to original score matrix
        score_matrix = np.array(orig_score_matrix, copy=True)

        oldI = np.zeros([length, length], dtype=np.int32)
        oldO = np.zeros([length, length], dtype=np.int32)
        curr_nodes = np.zeros([length], dtype=bool)
        reps = []

        for s in range(length):
            orig_score_matrix[s, s] = 0.0
            score_matrix[s, s] = 0.0
            curr_nodes[s] = True
            reps.append(set())
            reps[s].add(s)
            for t in range(s + 1, length):
                oldI[s, t] = s
                oldO[s, t] = t

                oldI[t, s] = t
                oldO[t, s] = s

        final_edges = dict()
        chuLiuEdmonds()
        par = np.zeros([max_length], np.int32)
        if labeled:
            type = np.ones([max_length], np.int32)
            type[0] = 0
        else:
            type = None

        for ch, pr in list(final_edges.items()):
            par[ch] = pr
            if labeled and ch != 0:
                type[ch] = label_id_matrix[pr, ch]

        par[0] = 0
        pars[i] = par
        if labeled:
            types[i] = type

    return pars, types



from typing import List
import plotly.graph_objects as go
from plotly.graph_objects import Figure
from plotly.subplots import make_subplots
import igraph

from aksara.conllu import ConlluData
from ._abstract_drawer import AbstractDrawer
from ._drawer_utils import (
    create_igraph,
    filter_multiword,
    get_conllu_root,
    is_in_ipython,
    create_igraph_layout
)

# pylint: disable=W0212, W0221
class PlotlyDrawer(AbstractDrawer):
    """
    Plotly-based dependency tree drawer

    """

    def draw(
            self,
            conllu_datas: List[List[ConlluData]],
            figure: Figure = None,
            fontsize: int = 12,
            width: int = None,
            height: int = None,
            scale_x: float = 1.0,
            scale_y: float = 1.0
        ) -> Figure:

        filtered_conllus = list(map(filter_multiword, conllu_datas))
        vertex_size = 4 * fontsize

        heights = list(map(lambda conllus: len(conllus) * vertex_size, filtered_conllus))

        if figure is None:
            figure = make_subplots(rows=len(conllu_datas), cols=1, row_heights=heights)

        total_subplot = self._get_total_subplots(figure)
        if total_subplot != len(conllu_datas):
            raise ValueError("Total subplot in `figure` must be equal to "
                             + "the lenght of `conllu_datas`")

        self.__add_traces_to_figure(conllu_datas, figure, fontsize, scale_x, scale_y)

        if height is None:
            height = 1.5 * sum(heights)

        figure.update_layout(
            showlegend=False,
            width=width,
            height=height,
            margin={'l':0, 'r':0, 'b':0, 't':0}
        )

        axis = {
            'showline':False,
            'zeroline':False,
            'showgrid':False,
            'showticklabels':False
        }
        figure.update_xaxes(axis)
        figure.update_yaxes(axis)

        if not is_in_ipython():
            figure.show()

        return figure

    def _get_total_subplots(self, figure: Figure) -> int:

        if figure is None:
            return 0
        if figure._grid_ref is None:
            return 1

        # created using make_subplot
        row_range, col_range = figure._get_subplot_rows_columns()
        return len(row_range) * len(col_range)

    def __add_traces_to_figure(
            self,
            conllu_datas: List[List[ConlluData]],
            figure: Figure,
            fontsize: int,
            scale_x: float,
            scale_y: float
        ) -> None:

        row = col = 1
        max_col = self.__get_total_columns(figure)

        for conllu_data in conllu_datas:

            graph = create_igraph(conllu_data)
            root = get_conllu_root(conllu_data)

            # use Reingold-Tilford Tree layout
            layout = create_igraph_layout(graph, root, scale_x, scale_y)

            edge_trace = self.__create_edge_trace(graph, layout)

            vertex_trace = self.__create_vertex_trace(graph, layout, fontsize)

            if figure._grid_ref is None:
                figure.add_trace(edge_trace)
                figure.add_trace(vertex_trace)
            else:
                figure.add_trace(edge_trace, row=row, col=col)
                figure.add_trace(vertex_trace, row=row, col=col)

            col += 1
            if col > max_col:
                row += 1
                col = 1

    def __get_total_columns(self, figure: Figure) -> int:
        max_col = 1
        total_subplot = self._get_total_subplots(figure)
        if total_subplot > 1:
            max_col = len(figure._get_subplot_rows_columns()[1])

        return max_col

    def __create_edge_trace(self, graph: igraph.Graph, layout: igraph.Layout):

        edge_list = [e.tuple for e in graph.es] # list of edges

        edge_xs = []
        edge_ys = []
        for (vertex1, vertex2) in edge_list:
            edge_xs += [layout.coords[vertex1][0], layout.coords[vertex2][0], None]
            edge_ys += [layout.coords[vertex1][1], layout.coords[vertex2][1], None]

        edge_trace = go.Scatter(
            x=edge_xs,
            y=edge_ys,
            mode='lines',
            line={'color':'rgb(210,210,210)', 'width':2},
            hoverinfo='none'
        )

        return edge_trace

    def __create_vertex_trace(self, graph: igraph.Graph, layout: igraph.Layout, fontsize: int):

        vertex_xs = []
        vertex_ys = []
        for x_coord, y_coord in layout.coords:
            vertex_xs.append(x_coord)
            vertex_ys.append(y_coord)

        names = self.__create_vertex_names(graph)

        vertex_size = 4 * fontsize

        vertex_trace = go.Scatter(
            x=vertex_xs,
            y=vertex_ys,
            mode='markers+text',
            marker={'symbol':'circle-dot',
                    'size':vertex_size,
                    'color':'#6175c1',
                    'line':{'color':'rgb(50,50,50)', 'width':2},
                    'opacity':0.2
            },
            text=names,
            textposition='middle center',
            textfont={'size':fontsize},
            hoverinfo='none'
        )

        return vertex_trace

    def __create_vertex_names(self, graph: igraph.Graph) -> List[str]:
        names = []
        for raw_name in graph.vs['name']:
            splitted_name = raw_name.split("\n")

            if len(splitted_name) > 2:
                splitted_name[1] = f'<span style="color:red">{splitted_name[1]}</span>'

            names.append('<br>'.join(splitted_name))
        return names

    def save_image(self,
            conllu_datas: List[List[ConlluData]],
            output_path: str,
            *args,
            **kwargs) -> None:

        figure = self.draw(conllu_datas, *args, **kwargs)
        figure.write_image(output_path)
        print(f"Image has been saved at {output_path}")



__author__ = 'max'

import pickle
import numpy as np
from gensim.models import KeyedVectors
import gzip

from .io import utils


def load_embedding_dict(embedding, embedding_path, normalize_digits=True):
    """
    load word embeddings from file
    :param embedding:
    :param embedding_path:
    :return: embedding dict, embedding dimention, caseless
    """
    print(("loading embedding: %s from %s" % (embedding, embedding_path)))
    if embedding == 'word2vec':
        # loading word2vec
        word2vec = KeyedVectors.load_word2vec_format(embedding_path, binary=False)
        embedd_dim = word2vec.vector_size
        return word2vec, embedd_dim
    elif embedding == 'glove':
        # loading GloVe
        embedd_dim = -1
        embedd_dict = dict()
        with gzip.open(embedding_path, 'r') as file:
            for line in file:
                line = line.strip()
                line = line.decode('utf-8')
                if len(line) == 0:
                    continue

                tokens = line.split()
                if embedd_dim < 0:
                    embedd_dim = len(tokens) - 1
                else:
                    assert (embedd_dim + 1 == len(tokens))
                embedd = np.empty([1, embedd_dim], dtype=np.float32)
                embedd[:] = tokens[1:]
                word = utils.DIGIT_RE.sub(b"0", tokens[0]) if normalize_digits else tokens[0]
                embedd_dict[word] = embedd
        return embedd_dict, embedd_dim
    elif embedding == 'senna':
        # loading Senna
        embedd_dim = -1
        embedd_dict = dict()
        with gzip.open(embedding_path, 'r') as file:
            for line in file:
                line = line.strip()
                line = line.decode('utf-8')
                if len(line) == 0:
                    continue

                tokens = line.split()
                if embedd_dim < 0:
                    embedd_dim = len(tokens) - 1
                else:
                    assert (embedd_dim + 1 == len(tokens))
                embedd = np.empty([1, embedd_dim], dtype=np.float32)
                embedd[:] = tokens[1:]
                word = utils.DIGIT_RE.sub(b"0", tokens[0]) if normalize_digits else tokens[0]
                embedd_dict[word] = embedd
        return embedd_dict, embedd_dim
    elif embedding == 'sskip':
        embedd_dim = -1
        embedd_dict = dict()
        with gzip.open(embedding_path, 'r') as file:
            # skip the first line
            file.readline()
            for line in file:
                line = line.strip()
                try:
                    line = line.decode('utf-8')
                    if len(line) == 0:
                        continue

                    tokens = line.split()
                    if len(tokens) < embedd_dim:
                        continue

                    if embedd_dim < 0:
                        embedd_dim = len(tokens) - 1

                    embedd = np.empty([1, embedd_dim], dtype=np.float32)
                    start = len(tokens) - embedd_dim
                    word = ' '.join(tokens[0:start])
                    embedd[:] = tokens[start:]
                    word = utils.DIGIT_RE.sub(b"0", word) if normalize_digits else word
                    embedd_dict[word] = embedd
                except UnicodeDecodeError:
                    continue
        return embedd_dict, embedd_dim
    elif embedding == 'polyglot':
        words, embeddings = pickle.load(open(embedding_path, 'rb'))
        _, embedd_dim = embeddings.shape
        embedd_dict = dict()
        for i, word in enumerate(words):
            embedd = np.empty([1, embedd_dim], dtype=np.float32)
            embedd[:] = embeddings[i, :]
            word = utils.DIGIT_RE.sub(b"0", word) if normalize_digits else word
            embedd_dict[word] = embedd
        return embedd_dict, embedd_dim

    else:
        raise ValueError("embedding should choose from [word2vec, senna, glove, sskip, polyglot]")



from typing import List
from abc import ABCMeta, abstractmethod

# pylint: disable=R0903
class AbstractTokenizer(metaclass=ABCMeta):
    """
    Abstract class for Tokenization.
    Subclass only need to implement tokenize method
    """

    @abstractmethod
    def tokenize(self, text: str, *args, **kwargs) -> List[str]:
        """
        Abstract method that performs tokenization

        Parameters
        ----------
        text: str
            text that will be tokenized
        *args : tuple
            see concrete implementation
        **kwargs: dict
            see concrete implementation
        """
        raise NotImplementedError("`tokenize` method is not implemented")



from abc import ABCMeta, abstractmethod
from typing import List
from aksara.conllu import ConlluData

class AbstractDrawer(metaclass=ABCMeta):
    """
    Abstract drawer class.
    Subclass must implement `draw` and `save_image` method
    
    """

    @abstractmethod
    def draw(self, conllu_datas: List[List[ConlluData]], *args, **kwargs):
        """
        abstract draw method
        
        Parameters
        ----------
        conllu_datas: list of list of ConlluData
            conllu_datas that will be drawn
        
        See Also
        --------
        aksara.dependency_tree._drawer.MatplotlibDrawer: Matplotlib drawer instance
        aksara.dependency_tree._drawer.PlotlyDrawer: Plotly drawer instance
        """

        raise NotImplementedError("`draw` method is not implemented")

    @abstractmethod
    def save_image(self, conllu_datas: List[List[ConlluData]], output_path: str, *args, **kwargs):
        """
        abstract draw method
        
        Parameters
        ----------
        conllu_datas: list of list of ConlluData
            conllu_datas that will be drawn and saved
        
        output_path: str
            file path at which the image will be saved
        
        See Also
        --------
        aksara.dependency_tree._drawer.MatplotlibDrawer: Matplotlib drawer instance
        aksara.dependency_tree._drawer.PlotlyDrawer: Plotly drawer instance
        """

        raise NotImplementedError("`save_image` method is not implemented")



from typing import List
import aksara.tokenizer as _internal_tokenizer

from .abstract_tokenizer import AbstractTokenizer

# pylint: disable=R0903
class BaseTokenizer(AbstractTokenizer):
    """Class to perform tokenization (multiword token will NOT be splitted)
    
    """

    def __init__(self) -> None:
        self.__base_tokenizer = _internal_tokenizer.BaseTokenizer()

    def tokenize(self, text: str, *args, **kwargs) -> List[str]:
        """
        tokenize `text`

        Parameters
        ----------

        text: str
            text that will be tokenized
        
        Returns
        -------
        list of str
            list of all token in text
        """

        # self.__base_tokenizer.tokenize return a tuple that contains word_list and sunflags
        word_list, _ = self.__base_tokenizer.tokenize(text)
        return word_list



"""This file contains various POS tagging functions"""

import os
from warnings import warn
from typing import List, Tuple, Literal
from aksara.core import analyze_sentence, split_sentence, sentences_from_file
from aksara.analyzer import BaseAnalyzer
from dependency_parsing.core import DependencyParser


class POSTagger:
    __all_input_modes = ["f", "s"]

    def __init__(self) -> None:
        self.analyzer = self.__get_default_analyzer()
        self.dependency_parser = self.__get_default_dependency_parser()

    def __get_sentences_string(self, input_str, input_mode, sep_regex) -> str:
        sentences = []
        if input_mode == "s":
            sentences = split_sentence(input_str, sep_regex)
        elif input_mode == "f":
            sentences = sentences_from_file(input_str, sep_regex)
        else:
            raise ValueError(
                f"input_mode must be one of {self.__all_input_modes}, but {input_mode} was given"
            )
        sentences_string = ""
        for i in sentences:
            sentences_string += i + " "

        return sentences_string

    def tag(
        self,
        input_src: str,
        input_mode: Literal["s", "f"] = "s",
        is_informal: bool = False,
        sep_regex: str = None,
    ) -> List[List[Tuple[str, str]]]:
        sentences_string = self.__get_sentences_string(input_src, input_mode, sep_regex)

        return self._pos_tag_multi_sentences(sentences_string, is_informal, sep_regex)

    def tag_to_file(
        self,
        input_src: str,
        write_path: str,
        input_mode: Literal["s", "f"] = "s",
        write_mode: Literal["x", "a", "w"] = "w",
        is_informal: bool = False,
        sep_regex: str = None,
    ) -> str:
        sentences_string = self.__get_sentences_string(input_src, input_mode, sep_regex)

        self._pos_tag_then_save_to_file(
            sentences_string, write_path, sep_regex, write_mode, is_informal
        )

        return write_path

    def _pos_tag_one_word(self, word: str, is_informal: bool = False) -> str:
        stripped_word = word.strip()

        if stripped_word == "":
            return ""

        # check space
        if stripped_word.find(" ") != -1:
            warn(f'expected a single word input but "{word}" was given')
            return "X"

        analyzed_word = analyze_sentence(
            stripped_word,
            self.analyzer,
            self.dependency_parser,
            v1=False,
            lemma=False,
            postag=True,
            informal=is_informal,
        )

        # check if tokenizer recognize stripped_word as a multi word
        if len(analyzed_word.split("\n")) > 1:
            warn(f'expected a single word input but "{word}" was given')
            return "X"
        _, _, tag = analyzed_word.split("\t")
        return tag

    def _pos_tag_one_sentence(
        self, sentence: str, is_informal: bool = False
    ) -> list[tuple[str, str]]:
        """
        performs pos tagging on the text that will be considered as a sentence,
        then returns a list of tuple containing each word with its corresponding
        POS tag as the result

        parameters
        ----------
        sentece: str
            the sentence that will be analyzed

        is_informal: bool
            tell aksara to treat text as informal one, default to False

        return
        ------
        ReturnType: list[tuple[str, str]]
            will return list of tuple containing each word with its corresponding POS tag
        """

        sentence = str(sentence).strip()

        if sentence == "":
            return []

        temp_result = analyze_sentence(
            text=sentence,
            analyzer=self.analyzer,
            dependency_parser=self.dependency_parser,
            v1=False,
            lemma=False,
            postag=True,
            informal=is_informal,
        )

        result = []

        for line in temp_result.split("\n"):
            _, word, postag = line.split("\t")
            result.append((word, postag))

        return result

    def _pos_tag_multi_sentences(
        self, sentences: str, is_informal: bool = False, sep_regex: str = None
    ) -> List[List[Tuple[str, str]]]:
        sentences = sentences.strip()

        if sentences == "":
            return []

        result: List[List[Tuple[str, str]]] = []

        sentence_list = split_sentence(sentences, sep_regex)
        for sentence in sentence_list:
            analyzed_sentence = self._pos_tag_one_sentence(sentence, is_informal)

            result.append(analyzed_sentence)

        return result

    def _pos_tag_then_save_to_file(
        self,
        text: str,
        file_path: str,
        sep_regex: str = None,
        write_mode: Literal["x", "a", "w"] = "w",
        is_informal: bool = False,
    ) -> bool:
        """
        performs pos tagging on the text, then save the result in the file specified by file_path

        parameters
        ----------
        text: str
            the text that will be analyzed

        file_path: str
            path to the file where the pos tagging result will be saved to

        sep_regex: str
            regex rule that specify end of sentence

        write_mode: str, ['x', 'a', 'w'], default to 'w'
            mode while writing on the file specified by file_path

            'x': create the specified file, throws error FileExistsError if already exists
            'a': append the pos tagging result at the end of the file,
                    create the specified file if not exists
            'w': overwrite the current file content with the pos tagging result,
                    create the specified file if not exists

            NOTE:
            - if write_mode not in ['x', 'a', 'w'] will throws ValueError
            - 'a' write_mode will add '\n\n' before the pos tagging result.
            - If you plan to use 'a' write_mode in a file that already has
                pos taging result on connlu format, the sent_id between
                the old (already in the file) and new pos tag result will not ne in sequence.

        is_informal: bool
            tell aksara to treat text as informal one, default to False

        return
        ------
        ReturnType: bool
            will return true if succesfully pos tag and save the result on the given file_path
        """

        all_write_modes = ["x", "a", "w"]

        if write_mode not in all_write_modes:
            raise ValueError(f"write_mode must be in {all_write_modes}")

        sentence_list = split_sentence(text, sep_regex)

        result_list: List[List[Tuple[str, str]]] = self._pos_tag_multi_sentences(
            text, sep_regex=sep_regex, is_informal=is_informal
        )

        header_format = "# sent_id = {}\n# text = {}"

        with open(file_path, write_mode, encoding="utf-8") as output_file:
            if write_mode == "a" and len(result_list) != 0:
                output_file.writelines("\n\n")

            for i, token_with_tag in enumerate(result_list):
                output_file.writelines(
                    header_format.format(str(i + 1), sentence_list[i])
                )

                for idx, (token, tag) in enumerate(token_with_tag):
                    output_file.writelines(f"\n{idx + 1}\t{token}\t{tag}")

                if i < len(result_list) - 1:  # don't add \n at the end of the file
                    output_file.writelines("\n\n")

        return True

    def __get_default_analyzer(self) -> BaseAnalyzer:
        bin_path = os.path.join(os.path.dirname(__file__), "bin", "aksara@v1.2.0.bin")

        return BaseAnalyzer(bin_path)

    def __get_default_dependency_parser(self) -> DependencyParser:
        return DependencyParser()



__author__ = 'max'


class CoNLL03Writer(object):
    def __init__(self, word_alphabet, char_alphabet, pos_alphabet, chunk_alphabet, ner_alphabet):
        self.__source_file = None
        self.__word_alphabet = word_alphabet
        self.__char_alphabet = char_alphabet
        self.__pos_alphabet = pos_alphabet
        self.__chunk_alphabet = chunk_alphabet
        self.__ner_alphabet = ner_alphabet

    def start(self, file_path):
        self.__source_file = open(file_path, 'w')

    def close(self):
        self.__source_file.close()

    def write(self, word, pos, chunk, predictions, targets, lengths):
        batch_size, _ = word.shape
        for i in range(batch_size):
            for j in range(lengths[i]):
                w = self.__word_alphabet.get_instance(word[i, j]).encode('utf-8')
                p = self.__pos_alphabet.get_instance(pos[i, j]).encode('utf-8')
                ch = self.__chunk_alphabet.get_instance(chunk[i, j]).encode('utf-8')
                tgt = self.__ner_alphabet.get_instance(targets[i, j]).encode('utf-8')
                pred = self.__ner_alphabet.get_instance(predictions[i, j]).encode('utf-8')
                self.__source_file.write('%d %s %s %s %s %s\n' % (j + 1, w, p, ch, tgt, pred))
            self.__source_file.write('\n')


class CoNLLXWriter(object):
    def __init__(self, word_alphabet, char_alphabet, pos_alphabet, type_alphabet):
        self.__source_file = None
        self.__word_alphabet = word_alphabet
        self.__char_alphabet = char_alphabet
        self.__pos_alphabet = pos_alphabet
        self.__type_alphabet = type_alphabet

    def start(self, file_path):
        self.__source_file = open(file_path, 'w')

    def close(self):
        self.__source_file.close()

    def write(self, word, pos, head, type, lengths, symbolic_root=False, symbolic_end=False):
        batch_size, _ = word.shape
        start = 1 if symbolic_root else 0
        end = 1 if symbolic_end else 0
        for i in range(batch_size):
            for j in range(start, lengths[i] - end):
                w = self.__word_alphabet.get_instance(word[i, j]).encode('utf-8')
                p = self.__pos_alphabet.get_instance(pos[i, j]).encode('utf-8')
                t = self.__type_alphabet.get_instance(type[i, j]).encode('utf-8')
                h = head[i, j]
                self.__source_file.write('%d\t%s\t_\t_\t%s\t_\t%d\t%s\n' % (j, w, p, h, t))
            self.__source_file.write('\n')



"""Base class for encoders and generic multi encoders."""



import torch.nn as nn

from .util_class import aeq


class EncoderBase(nn.Module):
    """
    Base encoder class. Specifies the interface used by different encoder types
    and required by :obj:`onmt.Models.NMTModel`.
    .. mermaid::
       graph BT
          A[Input]
          subgraph RNN
            C[Pos 1]
            D[Pos 2]
            E[Pos N]
          end
          F[Memory_Bank]
          G[Final]
          A-->C
          A-->D
          A-->E
          C-->F
          D-->F
          E-->F
          E-->G
    """

    def _check_args(self, src, lengths=None):
        n_batch, _, _ = src.size()
        if lengths is not None:
            n_batch_, = lengths.size()
            aeq(n_batch, n_batch_)

    def forward(self, src, lengths=None):
        """
        Args:
            src (:obj:`LongTensor`):
               padded sequences of sparse indices `[src_len x batch x nfeat]`
            lengths (:obj:`LongTensor`): length of each sequence `[batch]`
        Returns:
            (tuple of :obj:`FloatTensor`, :obj:`FloatTensor`):
                * final encoder state, used to initialize decoder
                * memory bank for attention, `[src_len x batch x hidden]`
        """
        raise NotImplementedError



import os
import json
import torch

from dependency_parsing.neuronlp2.io import conllx_data
from dependency_parsing.neuronlp2.io_multi import get_word_index_with_spec
from dependency_parsing.neuronlp2.models import BiRecurrentConvBiAffine
from dependency_parsing.model_proxy import ModelProxy

ROOT_CHAR = "_ROOT_CHAR"
ROOT_POS = "_ROOT_POS"
ROOT = "_ROOT"


class DependencyParser:
    def __init__(self, model_name='FR_GSD-ID_CSUI'):
        current_dir, _ = os.path.split(os.path.realpath(__file__))
        alphabet_path = os.path.join(current_dir, "alphabets")
        self.word_alphabet, self.char_alphabet, self.pos_alphabet, self.type_alphabet, _ = conllx_data.create_alphabets(
            alphabet_path,
            None, data_paths=[None, None], max_vocabulary_size=50000, embedd_dict=None)

        model_dir = os.path.join(current_dir, ".pretrained_model")
        if not os.path.isdir(model_dir):
            os.makedirs(model_dir)

        def load_model_arguments_from_json():
            arguments = json.load(open(arg_path, 'r'))
            return arguments['args'], arguments['kwargs']

        self.model_path = os.path.join(model_dir, model_name)
        self.model_path += '.pt'
        arg_path = self.model_path + '.arg.json'

        if not os.path.exists(self.model_path) or not os.path.exists(arg_path):
            ModelProxy.download_model(model_name)

        args, kwargs = load_model_arguments_from_json()
        self.model = BiRecurrentConvBiAffine(use_gpu=False, *args, **kwargs)

        self.model.load_state_dict(torch.load(self.model_path))

    def parse_rows(self, rows):
        modified_rows = rows.copy()
        heads_pred, types_pred = self.predict(rows)

        i = 1
        for row in modified_rows:
            if row[0].isnumeric():
                row[6] = str(heads_pred[i])
                row[7] = self.type_alphabet.get_instance(types_pred[i])
                i += 1
        return modified_rows

    def predict(self, rows):
        self.model.cpu()
        self.model.eval()

        words, chars, postags = self.convert_to_tensor(rows)
        mask = torch.tensor([[1 for i in range(words.shape[1])]])
        length = torch.tensor([words.shape[1]])

        temp_heads_pred, temp_types_pred = self.model.decode_mst(words, chars, postags, mask=mask, length=length,
                                                                 leading_symbolic=conllx_data.NUM_SYMBOLIC_TAGS)

        heads_pred = temp_heads_pred[0, :]
        types_pred = temp_types_pred[0, :]

        return heads_pred, types_pred

    def convert_to_tensor(self, rows):
        words = []
        chars = []
        postags = []
        max_char_len = 0

        for row in rows:
            if max_char_len < len(row[1]):
                max_char_len = len(row[1])

        words.append(self.word_alphabet.get_index(ROOT))
        temp_char = [self.char_alphabet.get_index(ROOT_CHAR)]
        for padding in range(max_char_len - 1):
            temp_char.append(1)
        chars.append(temp_char)
        postags.append(self.pos_alphabet.get_index(ROOT_POS))

        for row in rows:
            if row[0].isnumeric():
                words.append(get_word_index_with_spec(self.word_alphabet, row[1], 'id'))
                temp_char = []
                for char in row[1]:
                    temp_char.append(self.char_alphabet.get_index(char))
                for padding in range(max_char_len - len(row[1])):
                    temp_char.append(1)
                chars.append(temp_char)
                postags.append(self.pos_alphabet.get_index(row[3]))

        return torch.tensor([words]), torch.tensor([chars]), torch.tensor([postags])



__author__ = 'max'

from .instance import DependencyInstance, NERInstance
from .instance import Sentence
from .conllx_data import ROOT, ROOT_POS, ROOT_CHAR, ROOT_TYPE, END, END_POS, END_CHAR, END_TYPE
from . import utils
from ..io_multi import get_word_index_with_spec

class CoNLLXReader(object):
    def __init__(self, file_path, word_alphabet, char_alphabet, pos_alphabet, type_alphabet, lang_id=""):
        self.__source_file = open(file_path, 'r')
        self.__word_alphabet = word_alphabet
        self.__char_alphabet = char_alphabet
        self.__pos_alphabet = pos_alphabet
        self.__type_alphabet = type_alphabet
        self.lang_id = lang_id

    def close(self):
        self.__source_file.close()

    def getNext(self, normalize_digits=True, symbolic_root=False, symbolic_end=False):
        lines = []
        line = self.__source_file.readline()
        while line is not None and len(line.strip()) > 0:
            line = line.strip()
            lines.append(line.split('\t'))
            line = self.__source_file.readline()

        length = len(lines)
        if length == 0:
            return None

        words = []
        word_ids = []
        char_seqs = []
        char_id_seqs = []
        postags = []
        pos_ids = []
        types = []
        type_ids = []
        heads = []

        if symbolic_root:
            words.append(ROOT)
            word_ids.append(self.__word_alphabet.get_index(ROOT))
            char_seqs.append([ROOT_CHAR, ])
            char_id_seqs.append([self.__char_alphabet.get_index(ROOT_CHAR), ])
            postags.append(ROOT_POS)
            pos_ids.append(self.__pos_alphabet.get_index(ROOT_POS))
            types.append(ROOT_TYPE)
            type_ids.append(self.__type_alphabet.get_index(ROOT_TYPE))
            heads.append(0)

        for tokens in lines:
            chars = []
            char_ids = []
            for char in tokens[1]:
                chars.append(char)
                char_ids.append(self.__char_alphabet.get_index(char))
            if len(chars) > utils.MAX_CHAR_LENGTH:
                chars = chars[:utils.MAX_CHAR_LENGTH]
                char_ids = char_ids[:utils.MAX_CHAR_LENGTH]
            char_seqs.append(chars)
            char_id_seqs.append(char_ids)

            word = utils.DIGIT_RE.sub("0", tokens[1]) if normalize_digits else tokens[1]
            pos = tokens[4]
            head = int(tokens[6])
            type = utils.get_main_deplabel(tokens[7])

            words.append(word)
            # ===== modified: with lang_id prefix (with backoff to default lang)
            one_word_id = get_word_index_with_spec(self.__word_alphabet, word, self.lang_id)
            word_ids.append(one_word_id)
            # =====

            postags.append(pos)
            pos_ids.append(self.__pos_alphabet.get_index(pos))

            types.append(type)
            type_ids.append(self.__type_alphabet.get_index(type))

            heads.append(head)

        if symbolic_end:
            words.append(END)
            word_ids.append(self.__word_alphabet.get_index(END))
            char_seqs.append([END_CHAR, ])
            char_id_seqs.append([self.__char_alphabet.get_index(END_CHAR), ])
            postags.append(END_POS)
            pos_ids.append(self.__pos_alphabet.get_index(END_POS))
            types.append(END_TYPE)
            type_ids.append(self.__type_alphabet.get_index(END_TYPE))
            heads.append(0)

        return DependencyInstance(Sentence(words, word_ids, char_seqs, char_id_seqs), postags, pos_ids, heads, types,
                                  type_ids)


class CoNLL03Reader(object):
    def __init__(self, file_path, word_alphabet, char_alphabet, pos_alphabet, chunk_alphabet, ner_alphabet):
        self.__source_file = open(file_path, 'r')
        self.__word_alphabet = word_alphabet
        self.__char_alphabet = char_alphabet
        self.__pos_alphabet = pos_alphabet
        self.__chunk_alphabet = chunk_alphabet
        self.__ner_alphabet = ner_alphabet

    def close(self):
        self.__source_file.close()

    def getNext(self, normalize_digits=True):
        lines = []
        line = self.__source_file.readline()
        while line is not None and len(line.strip()) > 0:
            line = line.strip()
            line = line.decode('utf-8')
            lines.append(line.split(' '))
            line = self.__source_file.readline()

        length = len(lines)
        if length == 0:
            return None

        words = []
        word_ids = []
        char_seqs = []
        char_id_seqs = []
        postags = []
        pos_ids = []
        chunk_tags = []
        chunk_ids = []
        ner_tags = []
        ner_ids = []

        for tokens in lines:
            chars = []
            char_ids = []
            for char in tokens[1]:
                chars.append(char)
                char_ids.append(self.__char_alphabet.get_index(char))
            if len(chars) > utils.MAX_CHAR_LENGTH:
                chars = chars[:utils.MAX_CHAR_LENGTH]
                char_ids = char_ids[:utils.MAX_CHAR_LENGTH]
            char_seqs.append(chars)
            char_id_seqs.append(char_ids)

            word = utils.DIGIT_RE.sub(b"0", tokens[1]) if normalize_digits else tokens[1]
            pos = tokens[2]
            chunk = tokens[3]
            ner = tokens[4]

            words.append(word)
            word_ids.append(self.__word_alphabet.get_index(word))

            postags.append(pos)
            pos_ids.append(self.__pos_alphabet.get_index(pos))

            chunk_tags.append(chunk)
            chunk_ids.append(self.__chunk_alphabet.get_index(chunk))

            ner_tags.append(ner)
            ner_ids.append(self.__ner_alphabet.get_index(ner))

        return NERInstance(Sentence(words, word_ids, char_seqs, char_id_seqs), postags, pos_ids, chunk_tags, chunk_ids,
                           ner_tags, ner_ids)



"""
Position feed-forward network from "Attention is All You Need"
"""

import torch.nn as nn

from .util_class import LayerNorm


class PositionwiseFeedForward(nn.Module):
    """ A two-layer Feed-Forward-Network with residual layer norm.
        Args:
            d_model (int): the size of input for the first-layer of the FFN.
            d_ff (int): the hidden layer size of the second-layer
                              of the FNN.
            dropout (float): dropout probability(0-1.0).
    """

    def __init__(self, d_model, d_ff, relu_drop=0.1, res_drop=0.1):
        super(PositionwiseFeedForward, self).__init__()
        self.w_1 = nn.Linear(d_model, d_ff)
        self.w_2 = nn.Linear(d_ff, d_model)
        self.layer_norm = LayerNorm(d_model)
        self.relu_dropout = nn.Dropout(relu_drop)
        self.relu = nn.ReLU()
        self.residual_dropout = nn.Dropout(res_drop)

    def forward(self, x):
        """
        Layer definition.
        Args:
            input: [ batch_size, input_len, model_dim ]
        Returns:
            output: [ batch_size, input_len, model_dim ]
        """
        inter = self.relu_dropout(self.relu(self.w_1(self.layer_norm(x))))
        output = self.residual_dropout(self.w_2(inter))
        return output + x



import nltk
import os


class HMMLearn:
    N = 0
    TAGS = [
        'NOUN', 'PRON', 'VERB',
        'NUM', 'ADJ', 'ADP',
        'CCONJ', 'SCONJ', 'X',
        'AUX', 'DET', 'ADV',
        'PART', 'INTJ', 'PUNCT',
        'SYM', 'PROPN',
    ]

    def __init__(self, train_file=os.path.join("etc", "dataset", "preprocessed.txt"), trigram=False):
        self.trigram = trigram
        word_tags = []

        # fix path
        aksara_path = os.path.join(os.path.dirname(__file__), "..", "aksara")
        train_file = os.path.join(aksara_path, train_file)

        train = open(train_file, "r+", encoding="utf-8").readlines()
        for sentence in train:
            word_tags.append(("START", "START"))
            pairs = sentence.split(" ")
            pairs = pairs[:-1]  # ignore newline
            for pair in pairs:
                word = "/".join(pair.split("/")[0:-1])
                tag = pair.split("/")[-1]
                if tag != "":
                    word_tags.append((tag, word))
                else:
                    word_tags.append(("/", "PUNCT"))
                self.N += 1

            word_tags.append(("END", "END"))

        # get the conditional frequency distribution
        self.cfd_word_tags = nltk.ConditionalFreqDist(word_tags)

        tags = []

        for sentence in train:
            tags.append("START")
            pairs = sentence.split(" ")
            pairs = pairs[:-1]  # ignore newline
            for pair in pairs:
                tag = pair.split("/")[-1]
                if tag != "":
                    tags.append(tag)
                else:
                    tags.append("PUNCT")

            tags.append("END")

        # self.fd = nltk.FreqDist(tags)
        # for key, val in sorted(self.fd.items(), key=lambda x: x[1], reverse=True):
        #     print(key, val)

        if not trigram:
            self.cfd_tags = nltk.ConditionalFreqDist(nltk.bigrams(tags))
            # print(self.cfd_tags.keys())
        else:
            trigrams = [((x, y), z) for x, y, z in nltk.trigrams(tags)]
            self.cfd_tags = nltk.ConditionalFreqDist(trigrams)

    def get_emission_prob(self, word, tag, smoothing=True):
        num = self.cfd_word_tags[tag][word]
        denom = sum(list(self.cfd_word_tags[tag].values()))
        if smoothing:
            return (num + 1) / (denom + self.N)
        else:
            return num / denom

    def get_transition_prob(self, tag, prev_tag, smoothing=True):
        num = self.cfd_tags[prev_tag][tag]
        denom = sum(list(self.cfd_tags[prev_tag].values()))
        if smoothing:
            return (num + 1) / (denom + len(self.TAGS))
        else:
            return num / denom



__author__ = 'max'

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter


class BiAAttention(nn.Module):
    '''
    Bi-Affine attention layer.
    '''

    def __init__(self, input_size_encoder, input_size_decoder, num_labels, biaffine=True, **kwargs):
        '''

        Args:
            input_size_encoder: int
                the dimension of the encoder input.
            input_size_decoder: int
                the dimension of the decoder input.
            num_labels: int
                the number of labels of the crf layer
            biaffine: bool
                if apply bi-affine parameter.
            **kwargs:
        '''
        super(BiAAttention, self).__init__()
        self.input_size_encoder = input_size_encoder
        self.input_size_decoder = input_size_decoder
        self.num_labels = num_labels
        self.biaffine = biaffine

        self.W_d = Parameter(torch.Tensor(self.num_labels, self.input_size_decoder))
        self.W_e = Parameter(torch.Tensor(self.num_labels, self.input_size_encoder))
        self.b = Parameter(torch.Tensor(self.num_labels, 1, 1))
        if self.biaffine:
            self.U = Parameter(torch.Tensor(self.num_labels, self.input_size_decoder, self.input_size_encoder))
        else:
            self.register_parameter('U', None)

        self.reset_parameters()

    def reset_parameters(self):
        nn.init.xavier_uniform_(self.W_d)
        nn.init.xavier_uniform_(self.W_e)
        nn.init.constant_(self.b, 0.)
        if self.biaffine:
            nn.init.xavier_uniform_(self.U)

    def forward(self, input_d, input_e, mask_d=None, mask_e=None):
        '''

        Args:
            input_d: Tensor
                the decoder input tensor with shape = [batch, length_decoder, input_size]
            input_e: Tensor
                the child input tensor with shape = [batch, length_encoder, input_size]
            mask_d: Tensor or None
                the mask tensor for decoder with shape = [batch, length_decoder]
            mask_e: Tensor or None
                the mask tensor for encoder with shape = [batch, length_encoder]

        Returns: Tensor
            the energy tensor with shape = [batch, num_label, length, length]

        '''
        assert input_d.size(0) == input_e.size(0), 'batch sizes of encoder and decoder are requires to be equal.'
        batch, length_decoder, _ = input_d.size()
        _, length_encoder, _ = input_e.size()

        # compute decoder part: [num_label, input_size_decoder] * [batch, input_size_decoder, length_decoder]
        # the output shape is [batch, num_label, length_decoder]
        out_d = torch.matmul(self.W_d, input_d.transpose(1, 2)).unsqueeze(3)
        # compute decoder part: [num_label, input_size_encoder] * [batch, input_size_encoder, length_encoder]
        # the output shape is [batch, num_label, length_encoder]
        out_e = torch.matmul(self.W_e, input_e.transpose(1, 2)).unsqueeze(2)

        # output shape [batch, num_label, length_decoder, length_encoder]
        if self.biaffine:
            # compute bi-affine part
            # [batch, 1, length_decoder, input_size_decoder] * [num_labels, input_size_decoder, input_size_encoder]
            # output shape [batch, num_label, length_decoder, input_size_encoder]
            output = torch.matmul(input_d.unsqueeze(1), self.U)
            # [batch, num_label, length_decoder, input_size_encoder] * [batch, 1, input_size_encoder, length_encoder]
            # output shape [batch, num_label, length_decoder, length_encoder]
            output = torch.matmul(output, input_e.unsqueeze(1).transpose(2, 3))

            output = output + out_d + out_e + self.b
        else:
            output = out_d + out_d + self.b

        if mask_d is not None:
            output = output * mask_d.unsqueeze(1).unsqueeze(3) * mask_e.unsqueeze(1).unsqueeze(2)

        return output


class ConcatAttention(nn.Module):
    '''
    Concatenate attention layer.
    '''
    # TODO test it!

    def __init__(self, input_size_encoder, input_size_decoder, hidden_size, num_labels, **kwargs):
        '''

        Args:
            input_size_encoder: int
                the dimension of the encoder input.
            input_size_decoder: int
                the dimension of the decoder input.
            hidden_size: int
                the dimension of the hidden.
            num_labels: int
                the number of labels of the crf layer
            biaffine: bool
                if apply bi-affine parameter.
            **kwargs:
        '''
        super(ConcatAttention, self).__init__()
        self.input_size_encoder = input_size_encoder
        self.input_size_decoder = input_size_decoder
        self.hidden_size = hidden_size
        self.num_labels = num_labels

        self.W_d = Parameter(torch.Tensor(self.input_size_decoder, self.hidden_size))
        self.W_e = Parameter(torch.Tensor(self.input_size_encoder, self.hidden_size))
        self.b = Parameter(torch.Tensor(self.hidden_size))
        self.v = Parameter(torch.Tensor(self.hidden_size, self.num_labels))

        self.reset_parameters()

    def reset_parameters(self):
        nn.init.xavier_uniform(self.W_d)
        nn.init.xavier_uniform(self.W_e)
        nn.init.xavier_uniform(self.v)
        nn.init.constant(self.b, 0.)

    def forward(self, input_d, input_e, mask_d=None, mask_e=None):
        '''

        Args:
            input_d: Tensor
                the decoder input tensor with shape = [batch, length_decoder, input_size]
            input_e: Tensor
                the child input tensor with shape = [batch, length_encoder, input_size]
            mask_d: Tensor or None
                the mask tensor for decoder with shape = [batch, length_decoder]
            mask_e: Tensor or None
                the mask tensor for encoder with shape = [batch, length_encoder]

        Returns: Tensor
            the energy tensor with shape = [batch, num_label, length, length]

        '''
        assert input_d.size(0) == input_e.size(0), 'batch sizes of encoder and decoder are requires to be equal.'
        batch, length_decoder, _ = input_d.size()
        _, length_encoder, _ = input_e.size()

        # compute decoder part: [batch, length_decoder, input_size_decoder] * [input_size_decoder, hidden_size]
        # the output shape is [batch, length_decoder, hidden_size]
        # then --> [batch, 1, length_decoder, hidden_size]
        out_d = torch.matmul(input_d, self.W_d).unsqueeze(1)
        # compute decoder part: [batch, length_encoder, input_size_encoder] * [input_size_encoder, hidden_size]
        # the output shape is [batch, length_encoder, hidden_size]
        # then --> [batch, length_encoder, 1, hidden_size]
        out_e = torch.matmul(input_e, self.W_e).unsqueeze(2)

        # add them together [batch, length_encoder, length_decoder, hidden_size]
        out = F.tanh(out_d + out_e + self.b)

        # product with v
        # [batch, length_encoder, length_decoder, hidden_size] * [hidden, num_label]
        # [batch, length_encoder, length_decoder, num_labels]
        # then --> [batch, num_labels, length_decoder, length_encoder]
        return torch.matmul(out, self.v).transpose(1, 3)



__author__ = 'max'

import os.path
import random
import numpy as np
from .alphabet import Alphabet
from .logger import get_logger
from . import utils
import torch
from torch.autograd import Variable

# Special vocabulary symbols - we always put them at the start.
PAD = "_PAD"
PAD_POS = "_PAD_POS"
PAD_TYPE = "_<PAD>"
PAD_CHAR = "_PAD_CHAR"
PAD_CHUNK = "_PAD_CHUNK"
PAD_NER = "_PAD_NER"

ROOT = "_ROOT"
ROOT_POS = "_ROOT_POS"
ROOT_TYPE = "_<ROOT>"
ROOT_CHAR = "_ROOT_CHAR"

END = "_END"
END_POS = "_END_POS"
END_TYPE = "_<END>"
END_CHAR = "_END_CHAR"
_START_VOCAB = [PAD, ROOT, END]

UNK_ID = 0
PAD_ID_WORD = 1
PAD_ID_CHAR = 1
PAD_ID_TAG = 0

NUM_SYMBOLIC_TAGS = 3

_buckets = [10, 15, 20, 25, 30, 35, 40, 50, 60, 70, 80, 90, 100, 140]

from .reader import CoNLLXReader


def create_alphabets(alphabet_directory, train_path, data_paths=None, max_vocabulary_size=50000, embedd_dict=None,
                     min_occurence=1, normalize_digits=True):
    def expand_vocab():
        vocab_set = set(vocab_list)
        max_sent_length = 0
        for data_path in data_paths:
            # logger.info("Processing data: %s" % data_path)
            with open(data_path, 'r') as file:
                sent_length = 1
                for line in file:
                    line = line.decode('utf-8')
                    line = line.strip()
                    if len(line) == 0:
                        for val in _buckets:
                            if val > sent_length:
                                sent_length = val
                                break
                        if sent_length > max_sent_length:
                            max_sent_length = sent_length
                        sent_length = 1
                        continue

                    tokens = line.split('\t')
                    for char in tokens[1]:
                        char_alphabet.add(char)

                    word = utils.DIGIT_RE.sub(b"0", tokens[1]) if normalize_digits else tokens[1]
                    pos = tokens[4]
                    type = utils.get_main_deplabel(tokens[7])

                    pos_alphabet.add(pos)
                    type_alphabet.add(type)

                    if embedd_dict is not None:
                        if word not in vocab_set and (word in embedd_dict or word.lower() in embedd_dict):
                            vocab_set.add(word)
                            vocab_list.append(word)

        return max_sent_length

    def cosine_dict(dic1, dic2):
        numerator = 0
        dena = 0
        for key1, val1 in list(dic1.items()):
            numerator += val1 * dic2.get(key1, 0)
            dena += val1 * val1
        denb = 0
        for val2 in list(dic2.values()):
            denb += val2 * val2
        import math
        return numerator / math.sqrt(dena * denb)

    def check_sentences_for_types(data_path, src_types, freq_dict):
        """temporarily written for stat check - wasi"""
        # logger.info("Processing data: %s" % data_path)
        count1, count2 = 1, 0
        new_line = True
        type_alph_dict = dict()
        with open(data_path, 'r') as file:
            for line in file:
                line = line.decode('utf-8')
                line = line.strip()
                if len(line) == 0:
                    new_line = True
                    count1 += 1
                    continue

                tokens = line.split('\t')
                type = tokens[7]
                if type in type_alph_dict:
                    type_alph_dict[type] += 1
                else:
                    type_alph_dict[type] = 1

                if type not in src_types and new_line:
                    count2 += 1
                    new_line = False

            print(('total sentences = ', count1))
            print(('total sentences that contain a label not appeared in src lang = ', count2))
            print(('cosine similarity = ', cosine_dict(freq_dict, type_alph_dict)))

    logger = get_logger("Create Alphabets")
    word_alphabet = Alphabet('word', defualt_value=True, singleton=True)
    char_alphabet = Alphabet('character', defualt_value=True)
    pos_alphabet = Alphabet('pos')
    type_alphabet = Alphabet('type')
    max_sent_length = 0
    if not os.path.isdir(alphabet_directory):
        # logger.info("Creating Alphabets: %s" % alphabet_directory)

        char_alphabet.add(PAD_CHAR)
        pos_alphabet.add(PAD_POS)
        type_alphabet.add(PAD_TYPE)

        char_alphabet.add(ROOT_CHAR)
        pos_alphabet.add(ROOT_POS)
        type_alphabet.add(ROOT_TYPE)

        char_alphabet.add(END_CHAR)
        pos_alphabet.add(END_POS)
        type_alphabet.add(END_TYPE)

        vocab = dict()
        type_alph_dict = dict()  ## temporary
        sent_length = 1
        with open(train_path, 'r') as file:
            for line in file:
                line = line.decode('utf-8')
                line = line.strip()
                if len(line) == 0:
                    for val in _buckets:
                        if val > sent_length:
                            sent_length = val
                            break
                    if sent_length > max_sent_length:
                        max_sent_length = sent_length
                    sent_length = 1
                    continue

                tokens = line.split('\t')
                for char in tokens[1]:
                    char_alphabet.add(char)

                sent_length += 1
                word = utils.DIGIT_RE.sub(b"0", tokens[1]) if normalize_digits else tokens[1]
                pos = tokens[4]
                type = utils.get_main_deplabel(tokens[7])

                pos_alphabet.add(pos)
                type_alphabet.add(type)
                if type in type_alph_dict:
                    type_alph_dict[type] += 1
                else:
                    type_alph_dict[type] = 1

                if word in vocab:
                    vocab[word] += 1
                else:
                    vocab[word] = 1

        #############################
        # temporary code
        # if data_paths is not None:
        #     check_sentences_for_types(data_paths[1], type_alphabet.items(), type_alph_dict)
        ############################

        # collect singletons
        singletons = set([word for word, count in list(vocab.items()) if count <= min_occurence])

        # if a singleton is in pretrained embedding dict, set the count to min_occur + c
        if embedd_dict is not None:
            for word in list(vocab.keys()):
                if word in embedd_dict or word.lower() in embedd_dict:
                    vocab[word] += min_occurence

        vocab_list = _START_VOCAB + sorted(vocab, key=vocab.get, reverse=True)
        # logger.info("Total Vocabulary Size: %d" % len(vocab_list))
        # logger.info("Total Singleton Size:  %d" % len(singletons))
        vocab_list = [word for word in vocab_list if word in _START_VOCAB or vocab[word] > min_occurence]
        # logger.info("Total Vocabulary Size (w.o rare words): %d" % len(vocab_list))

        if len(vocab_list) > max_vocabulary_size:
            vocab_list = vocab_list[:max_vocabulary_size]

        if data_paths is not None:
            # logger.info("expanding vocab")
            sent_length = expand_vocab()
            if sent_length > max_sent_length:
                max_sent_length = sent_length

        # if data_paths is not None and embedd_dict is not None:
        #     expand_vocab()

        for word in vocab_list:
            word_alphabet.add(word)
            if word in singletons:
                word_alphabet.add_singleton(word_alphabet.get_index(word))

        word_alphabet.save(alphabet_directory)
        char_alphabet.save(alphabet_directory)
        pos_alphabet.save(alphabet_directory)
        type_alphabet.save(alphabet_directory)
    else:
        word_alphabet.load(alphabet_directory)
        char_alphabet.load(alphabet_directory)
        pos_alphabet.load(alphabet_directory)
        type_alphabet.load(alphabet_directory)

    word_alphabet.close()
    char_alphabet.close()
    pos_alphabet.close()
    type_alphabet.close()
    # logger.info("Word Alphabet Size (Singleton): %d (%d)" % (word_alphabet.size(), word_alphabet.singleton_size()))
    # logger.info("Character Alphabet Size: %d" % char_alphabet.size())
    # logger.info("POS Alphabet Size: %d" % pos_alphabet.size())
    # logger.info("Type Alphabet Size: %d" % type_alphabet.size())
    # logger.info("Maximum Sentence Length: %d" % max_sent_length)

    return word_alphabet, char_alphabet, pos_alphabet, type_alphabet, max_sent_length


def read_data(source_path, word_alphabet, char_alphabet, pos_alphabet, type_alphabet, max_size=None,
              normalize_digits=True, symbolic_root=False, symbolic_end=False, lang_id="", len_thresh=None):
    data = [[] for _ in _buckets]
    max_char_length = [0 for _ in _buckets]
    print(('Reading data from %s' % source_path))
    counter = 0
    counter_added = 0
    reader = CoNLLXReader(source_path, word_alphabet, char_alphabet, pos_alphabet, type_alphabet, lang_id=lang_id)
    inst = reader.getNext(normalize_digits=normalize_digits, symbolic_root=symbolic_root, symbolic_end=symbolic_end)
    while inst is not None and (not max_size or counter < max_size):
        counter += 1
        if counter % 10000 == 0:
            print(("reading data: %d" % counter))

        inst_size = inst.length()
        if inst_size <= len_thresh:
            sent = inst.sentence
            for bucket_id, bucket_size in enumerate(_buckets):
                if inst_size < bucket_size:
                    data[bucket_id].append([sent.word_ids, sent.char_id_seqs, inst.pos_ids, inst.heads, inst.type_ids])
                    max_len = max([len(char_seq) for char_seq in sent.char_seqs])
                    if max_char_length[bucket_id] < max_len:
                        max_char_length[bucket_id] = max_len
                    break
            counter_added += 1

        inst = reader.getNext(normalize_digits=normalize_digits, symbolic_root=symbolic_root, symbolic_end=symbolic_end)
    reader.close()
    print(("Total number of data: %d, used: %d" % (counter, counter_added)))
    return data, max_char_length


def get_batch(data, batch_size, word_alphabet=None, unk_replace=0.):
    data, max_char_length = data
    bucket_sizes = [len(data[b]) for b in range(len(_buckets))]
    total_size = float(sum(bucket_sizes))
    # A bucket scale is a list of increasing numbers from 0 to 1 that we'll use
    # to select a bucket. Length of [scale[i], scale[i+1]] is proportional to
    # the size if i-th training bucket, as used later.
    buckets_scale = [sum(bucket_sizes[:i + 1]) / total_size for i in range(len(bucket_sizes))]

    # Choose a bucket according to data distribution. We pick a random number
    # in [0, 1] and use the corresponding interval in train_buckets_scale.
    random_number = np.random.random_sample()
    bucket_id = min([i for i in range(len(buckets_scale)) if buckets_scale[i] > random_number])

    bucket_length = _buckets[bucket_id]
    char_length = min(utils.MAX_CHAR_LENGTH, max_char_length[bucket_id] + utils.NUM_CHAR_PAD)
    bucket_size = bucket_sizes[bucket_id]
    batch_size = min(bucket_size, batch_size)

    wid_inputs = np.empty([batch_size, bucket_length], dtype=np.int64)
    cid_inputs = np.empty([batch_size, bucket_length, char_length], dtype=np.int64)
    pid_inputs = np.empty([batch_size, bucket_length], dtype=np.int64)
    hid_inputs = np.empty([batch_size, bucket_length], dtype=np.int64)
    tid_inputs = np.empty([batch_size, bucket_length], dtype=np.int64)

    masks = np.zeros([batch_size, bucket_length], dtype=np.float32)
    single = np.zeros([batch_size, bucket_length], dtype=np.int64)

    for b in range(batch_size):
        wids, cid_seqs, pids, hids, tids = random.choice(data[bucket_id])

        inst_size = len(wids)
        # word ids
        wid_inputs[b, :inst_size] = wids
        wid_inputs[b, inst_size:] = PAD_ID_WORD
        for c, cids in enumerate(cid_seqs):
            cid_inputs[b, c, :len(cids)] = cids
            cid_inputs[b, c, len(cids):] = PAD_ID_CHAR
        cid_inputs[b, inst_size:, :] = PAD_ID_CHAR
        # pos ids
        pid_inputs[b, :inst_size] = pids
        pid_inputs[b, inst_size:] = PAD_ID_TAG
        # type ids
        tid_inputs[b, :inst_size] = tids
        tid_inputs[b, inst_size:] = PAD_ID_TAG
        # heads
        hid_inputs[b, :inst_size] = hids
        hid_inputs[b, inst_size:] = PAD_ID_TAG
        # masks
        masks[b, :inst_size] = 1.0

        if unk_replace:
            for j, wid in enumerate(wids):
                if word_alphabet.is_singleton(wid):
                    single[b, j] = 1

    if unk_replace:
        noise = np.random.binomial(1, unk_replace, size=[batch_size, bucket_length])
        wid_inputs = wid_inputs * (1 - noise * single)

    return wid_inputs, cid_inputs, pid_inputs, hid_inputs, tid_inputs, masks


def iterate_batch(data, batch_size, word_alphabet=None, unk_replace=0., shuffle=False):
    data, max_char_length = data
    bucket_sizes = [len(data[b]) for b in range(len(_buckets))]
    total_size = float(sum(bucket_sizes))
    bucket_indices = np.arange(len(_buckets))
    if shuffle:
        np.random.shuffle((bucket_indices))

    for bucket_id in bucket_indices:
        bucket_size = bucket_sizes[bucket_id]
        if bucket_size == 0:
            continue

        bucket_length = _buckets[bucket_id]
        char_length = min(utils.MAX_CHAR_LENGTH, max_char_length[bucket_id] + utils.NUM_CHAR_PAD)
        wid_inputs = np.empty([bucket_size, bucket_length], dtype=np.int64)
        cid_inputs = np.empty([bucket_size, bucket_length, char_length], dtype=np.int64)
        pid_inputs = np.empty([bucket_size, bucket_length], dtype=np.int64)
        hid_inputs = np.empty([bucket_size, bucket_length], dtype=np.int64)
        tid_inputs = np.empty([bucket_size, bucket_length], dtype=np.int64)

        masks = np.zeros([bucket_size, bucket_length], dtype=np.float32)
        single = np.zeros([bucket_size, bucket_length], dtype=np.int64)

        for i, inst in enumerate(data[bucket_id]):
            wids, cid_seqs, pids, hids, tids = inst
            inst_size = len(wids)
            # word ids
            wid_inputs[i, :inst_size] = wids
            wid_inputs[i, inst_size:] = PAD_ID_WORD
            for c, cids in enumerate(cid_seqs):
                cid_inputs[i, c, :len(cids)] = cids
                cid_inputs[i, c, len(cids):] = PAD_ID_CHAR
            cid_inputs[i, inst_size:, :] = PAD_ID_CHAR
            # pos ids
            pid_inputs[i, :inst_size] = pids
            pid_inputs[i, inst_size:] = PAD_ID_TAG
            # type ids
            tid_inputs[i, :inst_size] = tids
            tid_inputs[i, inst_size:] = PAD_ID_TAG
            # heads
            hid_inputs[i, :inst_size] = hids
            hid_inputs[i, inst_size:] = PAD_ID_TAG
            # masks
            masks[i, :inst_size] = 1.0
            if unk_replace:
                for j, wid in enumerate(wids):
                    if word_alphabet.is_singleton(wid):
                        single[i, j] = 1

        if unk_replace:
            noise = np.random.binomial(1, unk_replace, size=[bucket_size, bucket_length])
            wid_inputs = wid_inputs * (1 - noise * single)

        indices = None
        if shuffle:
            indices = np.arange(bucket_size)
            np.random.shuffle(indices)
        for start_idx in range(0, bucket_size, batch_size):
            if shuffle:
                excerpt = indices[start_idx:start_idx + batch_size]
            else:
                excerpt = slice(start_idx, start_idx + batch_size)
            yield wid_inputs[excerpt], cid_inputs[excerpt], pid_inputs[excerpt], hid_inputs[excerpt], \
                  tid_inputs[excerpt], masks[excerpt]


def read_data_to_variable(source_path, word_alphabet, char_alphabet, pos_alphabet, type_alphabet, max_size=None,
                          normalize_digits=True, symbolic_root=False, symbolic_end=False,
                          use_gpu=False, volatile=False, lang_id="", len_thresh=100000):
    data, max_char_length = read_data(source_path, word_alphabet, char_alphabet, pos_alphabet, type_alphabet,
                                      max_size=max_size, normalize_digits=normalize_digits, symbolic_root=symbolic_root,
                                      symbolic_end=symbolic_end, lang_id=lang_id, len_thresh=len_thresh)
    bucket_sizes = [len(data[b]) for b in range(len(_buckets))]

    data_variable = []

    for bucket_id in range(len(_buckets)):
        bucket_size = bucket_sizes[bucket_id]
        if bucket_size == 0:
            data_variable.append((1, 1))
            continue

        bucket_length = _buckets[bucket_id]
        char_length = min(utils.MAX_CHAR_LENGTH, max_char_length[bucket_id] + utils.NUM_CHAR_PAD)
        wid_inputs = np.empty([bucket_size, bucket_length], dtype=np.int64)
        cid_inputs = np.empty([bucket_size, bucket_length, char_length], dtype=np.int64)
        pid_inputs = np.empty([bucket_size, bucket_length], dtype=np.int64)
        hid_inputs = np.empty([bucket_size, bucket_length], dtype=np.int64)
        tid_inputs = np.empty([bucket_size, bucket_length], dtype=np.int64)

        masks = np.zeros([bucket_size, bucket_length], dtype=np.float32)
        single = np.zeros([bucket_size, bucket_length], dtype=np.int64)

        lengths = np.empty(bucket_size, dtype=np.int64)

        for i, inst in enumerate(data[bucket_id]):
            wids, cid_seqs, pids, hids, tids = inst
            inst_size = len(wids)
            lengths[i] = inst_size
            # word ids
            wid_inputs[i, :inst_size] = wids
            wid_inputs[i, inst_size:] = PAD_ID_WORD
            for c, cids in enumerate(cid_seqs):
                cid_inputs[i, c, :len(cids)] = cids
                cid_inputs[i, c, len(cids):] = PAD_ID_CHAR
            cid_inputs[i, inst_size:, :] = PAD_ID_CHAR
            # pos ids
            pid_inputs[i, :inst_size] = pids
            pid_inputs[i, inst_size:] = PAD_ID_TAG
            # type ids
            tid_inputs[i, :inst_size] = tids
            tid_inputs[i, inst_size:] = PAD_ID_TAG
            # heads
            hid_inputs[i, :inst_size] = hids
            hid_inputs[i, inst_size:] = PAD_ID_TAG
            # masks
            masks[i, :inst_size] = 1.0
            for j, wid in enumerate(wids):
                if word_alphabet.is_singleton(wid):
                    single[i, j] = 1

        with torch.no_grad():
            words = torch.from_numpy(wid_inputs)
            chars = torch.from_numpy(cid_inputs)
            pos = torch.from_numpy(pid_inputs)
            heads = torch.from_numpy(hid_inputs)
            types = torch.from_numpy(tid_inputs)
            masks = torch.from_numpy(masks)
            single = torch.from_numpy(single)
            lengths = torch.from_numpy(lengths)
            if use_gpu:
                words = words.cuda()
                chars = chars.cuda()
                pos = pos.cuda()
                heads = heads.cuda()
                types = types.cuda()
                masks = masks.cuda()
                single = single.cuda()
                lengths = lengths.cuda()

        data_variable.append((words, chars, pos, heads, types, masks, single, lengths))

    return data_variable, bucket_sizes


def get_batch_variable(data, batch_size, unk_replace=0.):
    data_variable, bucket_sizes = data
    total_size = float(sum(bucket_sizes))
    # A bucket scale is a list of increasing numbers from 0 to 1 that we'll use
    # to select a bucket. Length of [scale[i], scale[i+1]] is proportional to
    # the size if i-th training bucket, as used later.
    buckets_scale = [sum(bucket_sizes[:i + 1]) / total_size for i in range(len(bucket_sizes))]

    # Choose a bucket according to data distribution. We pick a random number
    # in [0, 1] and use the corresponding interval in train_buckets_scale.
    random_number = np.random.random_sample()
    bucket_id = min([i for i in range(len(buckets_scale)) if buckets_scale[i] > random_number])
    bucket_length = _buckets[bucket_id]

    words, chars, pos, heads, types, masks, single, lengths = data_variable[bucket_id]
    bucket_size = bucket_sizes[bucket_id]
    batch_size = min(bucket_size, batch_size)
    index = torch.randperm(bucket_size).long()[:batch_size]
    if words.is_cuda:
        index = index.cuda()

    words = words[index]
    if unk_replace:
        ones = Variable(single.data.new(batch_size, bucket_length).fill_(1))
        noise = Variable(masks.data.new(batch_size, bucket_length).bernoulli_(unk_replace).long())
        words = words * (ones - single[index] * noise)

    return words, chars[index], pos[index], heads[index], types[index], masks[index], lengths[index]


def iterate_batch_variable(data, batch_size, unk_replace=0., shuffle=False):
    data_variable, bucket_sizes = data

    bucket_indices = np.arange(len(_buckets))
    if shuffle:
        np.random.shuffle((bucket_indices))

    for bucket_id in bucket_indices:
        bucket_size = bucket_sizes[bucket_id]
        bucket_length = _buckets[bucket_id]
        if bucket_size == 0:
            continue

        words, chars, pos, heads, types, masks, single, lengths = data_variable[bucket_id]
        if unk_replace:
            ones = Variable(single.data.new(bucket_size, bucket_length).fill_(1))
            noise = Variable(masks.data.new(bucket_size, bucket_length).bernoulli_(unk_replace).long())
            words = words * (ones - single * noise)

        indices = None
        if shuffle:
            indices = torch.randperm(bucket_size).long()
            if words.is_cuda:
                indices = indices.cuda()
        for start_idx in range(0, bucket_size, batch_size):
            if shuffle:
                excerpt = indices[start_idx:start_idx + batch_size]
            else:
                excerpt = slice(start_idx, start_idx + batch_size)
            yield words[excerpt], chars[excerpt], pos[excerpt], heads[excerpt], types[excerpt], \
                  masks[excerpt], lengths[excerpt]



#!/usr/bin/python3

import stat
from sys import platform
from tempfile import NamedTemporaryFile

import os
import re
import subprocess


# BIN_FILE = "bin/umabi@v1.0.3.bin"


class BaseAnalyzer:

    def __init__(self, bin_file):
        self.__bin_file = bin_file

    def __get_analysis(self, word):
        auto_delate = True
        if platform == "win32":
            auto_delate = False

        temp_file = NamedTemporaryFile(delete=auto_delate)
        with open(temp_file.name, 'w', encoding="utf-8") as f:
            f.write("load " + self.__bin_file + "\n")
            f.write("apply up " + word)

        os.chmod(temp_file.name, 777)
        temp_file.file.close()
        out = subprocess.check_output(['foma', '-q', '-f', temp_file.name])

        if platform == "win32":
            # temp file in windows is default to READ_ONLY
            # os.unlink will raise error on READ_ONLY file
            os.chmod(temp_file.name, stat.S_IWRITE)
            os.unlink(temp_file.name)
        return repr(out)[2:-1]

    def analyze(self, word):
        # Get lemma from Foma
        analysis = self.__get_analysis(word)
        analysis = analysis[:-2]  # Remove most right \n

        if "@informal" == analysis[:9]:
            analysis = analysis[9:]

        if analysis == '???':
            if "@informal" == word[:9]:
                word = word[9:]
            analysis = self.__analyze_unknown(word)

        analysis = list(set(analysis.split("\\n")))
        return "\\n".join(analysis)

    def __trim_analysis(self, analysis):
        # Remove the clitics
        temp = analysis.split("+_")[-1]  # Remove proclitic
        temp = temp.split("_+")[0]  # Remove enclitic
        return temp.split("+")

    def __get_postag(self, text):
        return self.__trim_analysis(text)[1]

    def __get_lemma(self, text):
        return self.__trim_analysis(text)[0]

    def __analyze_redup(self, surface):
        # Regex pattern
        redup_pattern = r'^([a-z]+)(\-)([a-z]+)$'

        # Setting up
        redup_search = re.search(redup_pattern, surface, re.IGNORECASE)
        if not redup_search:
            return "???"
        first_word = redup_search.group(1)
        second_word = redup_search.group(3)

        # Get analysis for each word
        first_word_analysis = self.__get_analysis(first_word)[:-2]
        second_word_analysis = self.__get_analysis(second_word)[:-2]

        if first_word_analysis == "???":
            return "???"

        # Write up results
        new_analysis = ""
        new_postag = ""
        if self.__get_lemma(first_word_analysis) == self.__get_lemma(second_word_analysis):
            new_postag = self.__get_postag(first_word_analysis)
            new_analysis = first_word_analysis
        elif second_word_analysis == "???":
            new_postag = self.__get_postag(first_word_analysis)
            new_analysis = first_word_analysis
        else:
            return "???"

        if new_postag == "NOUN":
            new_analysis = re.sub(r'(?<=\+Number=)Sing', 'Plur', new_analysis)

        return new_analysis

    def __analyze_unknown(self, surface):
        # Regex pattern
        redup_pattern = re.compile(r'([a-z]+)(\-)([a-z]+)')
        proper_noun_pattern = re.compile(r'[A-Z]+[a-z]*')
        sym_pattern = re.compile(
            r'[^\w“”,.?!()—":\'(\-\-)\-]|[\w\-\.]+@([\w\-]+\.)+[\w\-]{2,4}|:[\S](?=\s|$)|:-[\S](?=\s|$)')
        punct_pattern = re.compile(r'[“”,.?!()—":\'(\-\-)\-]')

        # Word list
        proper_noun_lst = ['of', 'the', "n't", "'s", "'m"]

        # Check every pattern
        analysis = "???"

        if redup_pattern.match(surface):
            analysis = self.__analyze_redup(surface)

        if analysis != "???":
            return analysis

        postag = "X"
        if proper_noun_pattern.match(surface):
            postag = 'PROPN'
        elif surface in proper_noun_lst:
            postag = 'PROPN'
        elif sym_pattern.match(surface):
            postag = "SYM"
        elif punct_pattern.match(surface):
            surface = punct_pattern.match(surface).group(0)
            postag = "PUNCT"

        analysis = "".join([surface, "+", postag])
        analysis += self.__get_feature_tags(analysis, postag)
        return analysis

    def __get_feature_tags(self, analysis, postag):
        tags = []

        if postag == "X":
            tags.append("Foreign=Yes")

        # Add first plus sign
        tags = "+".join(sorted(tags))
        if tags:
            tags = "+" + tags

        return tags



from typing import List, Union
import igraph
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import numpy as np

from aksara.conllu import ConlluData
from ._abstract_drawer import AbstractDrawer
from ._drawer_utils import create_igraph, filter_multiword, get_conllu_root, create_igraph_layout

# pylint: disable=W0212, W0221
class MatplotlibDrawer(AbstractDrawer):
    """
    Matplotlib-based dependency tree drawer

    """

    def draw(
            self,
            conllu_datas: List[List[ConlluData]],
            axs: Union[List[plt.Axes], plt.Axes, np.ndarray] = None,
            scale_x: float = 1.0,
            scale_y: float = 1.0
        ) -> Figure:

        figure = None
        if axs is None:
            _, axs = plt.subplots(len(conllu_datas))

        if isinstance(axs, plt.Axes):
            axs = [axs]

        if isinstance(axs, np.ndarray) and len(axs.shape) > 1:
            raise RuntimeError("axs only accepts one dimensional array."
                             + "\n\tIf axs has been created using `plt.subplots`,"
                             + "then set ncols to 1"
                             + "\n\t(e.g. fig, axs = plt.subplots(n, 1))")

        if len(axs) != len(conllu_datas):
            raise RuntimeError("`axs` length must match `conllu_datas` length")

        figure = axs[0].figure

        for axis in axs:
            axis.clear()

        for i, conllu_data in enumerate(conllu_datas):
            filtered_conllu = filter_multiword(conllu_data)
            graph = create_igraph(filtered_conllu)
            root = get_conllu_root(filtered_conllu)

            graph.vs["color"] = [(0, 0, 0, 0.3)] * len(filtered_conllu)

            # use Reingold-Tilford Tree layout
            layout = create_igraph_layout(graph, root, scale_x, scale_y)

            # Currently, igraph support for matplotlib annotation is limited
            for j, coords in enumerate(layout):
                axs[i].annotate(
                    graph.vs["name"][j],
                    coords,
                    ha='left',
                    va='bottom',
                    bbox={
                        "boxstyle":'round', 
                        "facecolor":'wheat'
                    }
                )

            igraph.plot(graph, layout=layout, target=axs[i])

        if not plt.isinteractive():
            plt.show()

        plt.close(figure)

        return figure

    def save_image(self,
            conllu_datas: List[List[ConlluData]],
            output_path: str,
            *args,
            **kwargs) -> None:

        figure = self.draw(conllu_datas, *args, **kwargs)
        figure.savefig(output_path)

        print(f"Dependency tree image is successfully saved at {output_path}")



import os

from typing import List, Literal
from aksara.core import analyze_sentence, split_sentence, sentences_from_file
from aksara.analyzer import BaseAnalyzer
from dependency_parsing.core import DependencyParser

class MorphologicalAnalyzer:
    """
    Class for aksara morphological analyzer feature
    """

    __all_input_modes = ["f", "s"]

    def __init__(self):
        self.default_analyzer = self.__get_default_analyzer()
        self.default_dependency_parser = self.__get_default_dependency_parser()

    def __get_sentence_list(self, input_str, input_mode, sep_regex) -> List[str]:
        if input_mode == "s":
            return split_sentence(input_str, sep_regex)

        if input_mode == "f":
            return sentences_from_file(input_str, sep_regex)

        raise ValueError(
            f"input_mode must be one of {self.__all_input_modes}, but {input_mode} was given"
        )

    def analyze(
        self, input_src: str,
        input_mode: Literal["f", "s"] = "s",
        is_informal: bool = False,
        sep_regex: str = None,
    ) -> List[List[tuple[str, List]]]:

        sentence_list = self.__get_sentence_list(input_src.strip(), input_mode, sep_regex)

        if len(sentence_list) == 1 and sentence_list[0] == "":
            return []

        result = []

        for sentence in sentence_list:
            sentence_result = self._analyze_one_sentence(
                sentence, is_informal
            )
            result.append(sentence_result)

        return result

    def analyze_to_file(
            self, input_src: str,
            write_path: str,
            input_mode: Literal["f", "s"] = "s",
            write_mode: Literal['a', 'w', 'x'] = 'x',
            is_informal: bool = False,
            sep_regex: str = None
    ) -> str:

        all_write_modes = ["x", "a", "w"]

        if write_mode not in all_write_modes:
            raise ValueError(f"write_mode must be in {all_write_modes}")

        sentence_list = self.__get_sentence_list(input_src.strip(), input_mode, sep_regex)
        analyzed_text = self.analyze(input_src, input_mode, is_informal, sep_regex)

        with open(write_path, write_mode, encoding="utf-8") as output_file:
            if write_mode == "a" and len(analyzed_text) != 0:
                output_file.writelines("\n\n")

            for i, sentence_feat in enumerate(analyzed_text):
                output_file.writelines(
                    f"# sent_id = {str(i + 1)}\n# text = {sentence_list[i]}")

                for idx, (form, feat) in enumerate(sentence_feat):
                    combined_feat = "_"
                    if feat != []:
                        combined_feat = "|".join(feat)

                    output_file.writelines(f"\n{idx + 1}\t{form}\t{combined_feat}")

                if i < len(analyzed_text) - 1:  # don't add \n at the end of the file
                    output_file.writelines("\n\n")

        return os.path.realpath(write_path)

    def _analyze_one_sentence(
        self, sentence: str,
        is_informal: bool = False
    ) -> List[tuple[str, List]]:

        sentence = sentence.strip()

        if sentence == "":
            return []

        analyzed_sentence = analyze_sentence(
            sentence,
            self.default_analyzer,
            self.default_dependency_parser,
            v1=False,
            lemma=False,
            postag=False,
            informal=is_informal
        )

        result = []
        for row in analyzed_sentence.split("\n"):
            _, form, _, _, _, feat, _, _, _, _ = row.split("\t")

            if feat != "_":
                feat = feat.split("|")
                result.append((form, feat))
            else:
                result.append((form, []))

        return result

    def __get_default_analyzer(self) -> BaseAnalyzer:
        bin_path = os.path.join(os.path.dirname(__file__), "bin", "aksara@v1.2.0.bin")
        return BaseAnalyzer(bin_path)

    def __get_default_dependency_parser(self) -> DependencyParser:
        return DependencyParser()



from .analyzer import (
    BaseAnalyzer,
)

from .tokenizer import (
    BaseTokenizer,
)

from .formatter import (
    to_conllu_line,
    to_conllu_line_with_range,
)

from .core import (
    analyze_sentence,
    create_args_parser,
    get_num_lines,
    split_sentence,
)



""" Misc classes """
import torch
import torch.nn as nn


class LayerNorm(nn.Module):
    """
        Layer Normalization class
    """

    def __init__(self, features, eps=1e-6):
        super(LayerNorm, self).__init__()
        self.a_2 = nn.Parameter(torch.ones(features))
        self.b_2 = nn.Parameter(torch.zeros(features))
        self.eps = eps

    def forward(self, x):
        mean = x.mean(-1, keepdim=True)
        std = x.std(-1, keepdim=True)
        return self.a_2 * (x - mean) / (std + self.eps) + self.b_2


def aeq(*args):
    """
    Assert all arguments have the same value
    """
    arguments = (arg for arg in args)
    first = next(arguments)
    assert all(arg == first for arg in arguments), \
        "Not all arguments have the same value: " + str(args)



from typing import List
import os

from dependency_parsing.core import DependencyParser
from ..core import analyze_sentence
from ..analyzer import BaseAnalyzer

from .abstract_tokenizer import AbstractTokenizer

# pylint: disable=R0903
class MultiwordTokenizer(AbstractTokenizer):
    """Class to perform tokenization (multiword token will be splitted)
    
    """

    def __init__(self) -> None:
        __bin_path = os.path.join(os.path.dirname(__file__), "..", "bin", "aksara@v1.2.0.bin")
        self.__base_analyzer = BaseAnalyzer(__bin_path)
        self.__dependency_parser = DependencyParser()

    def tokenize(self, text: str, *args, **kwargs) -> List[str]:
        """
        tokenize `text`

        Parameters
        ----------

        text: str
            text that will be tokenized
        
        Returns
        -------
        list of str
            list of all token in text
        """

        stripped_sentence = text.strip()

        if stripped_sentence == "":
            return []

        analyzed_result = analyze_sentence(
            stripped_sentence,
            self.__base_analyzer,
            self.__dependency_parser,
            informal=True,
            v1=False,
            postag=True,
            lemma=False
        )

        result: List[str] = []
        for conllu_row in analyzed_result.split("\n"):
            idx, form, _ = conllu_row.split("\t")
            if "-" not in idx:
                result.append(form)

        return result



__author__ = 'max'

import math
import torch
import torch.nn as nn
from torch.nn.parameter import Parameter
from torch.autograd import Variable

def default_initializer(hidden_size):
    stdv = 1.0 / math.sqrt(hidden_size)
    def forward(tensor):
        nn.init.uniform(tensor, -stdv, stdv)

    return forward

class VarMaskedRNNBase(nn.Module):
    def __init__(self, Cell, input_size, hidden_size,
                 num_layers=1, bias=True, batch_first=False,
                 dropout=(0, 0), bidirectional=False, initializer=None, **kwargs):

        super(VarMaskedRNNBase, self).__init__()
        self.Cell = Cell
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.bias = bias
        self.batch_first = batch_first
        self.bidirectional = bidirectional
        self.lstm = False
        num_directions = 2 if bidirectional else 1

        self.all_cells = []
        for layer in range(num_layers):
            for direction in range(num_directions):
                layer_input_size = input_size if layer == 0 else hidden_size * num_directions

                cell = self.Cell(layer_input_size, hidden_size, self.bias, p=dropout, initializer=initializer, **kwargs)
                self.all_cells.append(cell)
                self.add_module('cell%d' % (layer * num_directions + direction), cell)

    def reset_parameters(self):
        for cell in self.all_cells:
            cell.reset_parameters()

    def reset_noise(self, batch_size):
        for cell in self.all_cells:
            cell.reset_noise(batch_size)

    def forward(self, input, mask=None, hx=None):
        batch_size = input.size(0) if self.batch_first else input.size(1)
        if hx is None:
            num_directions = 2 if self.bidirectional else 1
            hx = torch.autograd.Variable(input.data.new(self.num_layers * num_directions, batch_size, self.hidden_size).zero_())
            if self.lstm:
                hx = (hx, hx)

        func = rnn_F.AutogradVarMaskedRNN(num_layers=self.num_layers,
                                          batch_first=self.batch_first,
                                          bidirectional=self.bidirectional,
                                          lstm=self.lstm)

        self.reset_noise(batch_size)

        output, hidden = func(input, self.all_cells, hx, None if mask is None else mask.view(mask.size() + (1,)))
        return output, hidden

    def step(self, input, hx=None, mask=None):
        '''
        execute one step forward (only for one-directional RNN).
        Args:
            input (batch, input_size): input tensor of this step.
            hx (num_layers, batch, hidden_size): the hidden state of last step.
            mask (batch): the mask tensor of this step.

        Returns:
            output (batch, hidden_size): tensor containing the output of this step from the last layer of RNN.
            hn (num_layers, batch, hidden_size): tensor containing the hidden state of this step
        '''
        assert not self.bidirectional, "step only cannot be applied to bidirectional RNN."
        batch_size = input.size(0)
        if hx is None:
            hx = torch.autograd.Variable(input.data.new(self.num_layers, batch_size, self.hidden_size).zero_())
            if self.lstm:
                hx = (hx, hx)

        func = rnn_F.AutogradVarMaskedStep(num_layers=self.num_layers, lstm=self.lstm)

        output, hidden = func(input, self.all_cells, hx, mask)
        return output, hidden

class VarMaskedFastLSTM(VarMaskedRNNBase):
    r"""Applies a multi-layer long short-term memory (LSTM) RNN to an input
    sequence.


    For each element in the input sequence, each layer computes the following
    function:

    .. math::

            \begin{array}{ll}
            i_t = \mathrm{sigmoid}(W_{ii} x_t + b_{ii} + W_{hi} h_{(t-1)} + b_{hi}) \\
            f_t = \mathrm{sigmoid}(W_{if} x_t + b_{if} + W_{hf} h_{(t-1)} + b_{hf}) \\
            g_t = \tanh(W_{ig} x_t + b_{ig} + W_{hc} h_{(t-1)} + b_{hg}) \\
            o_t = \mathrm{sigmoid}(W_{io} x_t + b_{io} + W_{ho} h_{(t-1)} + b_{ho}) \\
            c_t = f_t * c_{(t-1)} + i_t * g_t \\
            h_t = o_t * \tanh(c_t)
            \end{array}

    where :math:`h_t` is the hidden state at time `t`, :math:`c_t` is the cell
    state at time `t`, :math:`x_t` is the hidden state of the previous layer at
    time `t` or :math:`input_t` for the first layer, and :math:`i_t`,
    :math:`f_t`, :math:`g_t`, :math:`o_t` are the input, forget, cell,
    and out gates, respectively.

    Args:
        input_size: The number of expected features in the input x
        hidden_size: The number of features in the hidden state h
        num_layers: Number of recurrent layers.
        bias: If False, then the layer does not use bias weights b_ih and b_hh.
            Default: True
        batch_first: If True, then the input and output tensors are provided
            as (batch, seq, feature)
        dropout: (dropout_in, dropout_hidden) tuple.
            If non-zero, introduces a dropout layer on the input and hidden of the each
            RNN layer with dropout rate dropout_in and dropout_hidden, resp.
        bidirectional: If True, becomes a bidirectional RNN. Default: False

    Inputs: input, mask, (h_0, c_0)
        - **input** (seq_len, batch, input_size): tensor containing the features
          of the input sequence.
          **mask** (seq_len, batch): 0-1 tensor containing the mask of the input sequence.
        - **h_0** (num_layers \* num_directions, batch, hidden_size): tensor
          containing the initial hidden state for each element in the batch.
        - **c_0** (num_layers \* num_directions, batch, hidden_size): tensor
          containing the initial cell state for each element in the batch.


    Outputs: output, (h_n, c_n)
        - **output** (seq_len, batch, hidden_size * num_directions): tensor
          containing the output features `(h_t)` from the last layer of the RNN,
          for each t. If a :class:`torch.nn.utils.rnn.PackedSequence` has been
          given as the input, the output will also be a packed sequence.
        - **h_n** (num_layers * num_directions, batch, hidden_size): tensor
          containing the hidden state for t=seq_len
        - **c_n** (num_layers * num_directions, batch, hidden_size): tensor
          containing the cell state for t=seq_len
    """

    def __init__(self, *args, **kwargs):
        super(VarMaskedFastLSTM, self).__init__(VarFastLSTMCell, *args, **kwargs)
        self.lstm = True



#################################################################################
# Class ConlluData
#
# represents a token (a row of 10 columns) in CoNLLU format
#
# Author: Ika Alfina (ika.alfina@cs.ui.ac.id)
#
#################################################################################

"""
Class Wrapper for ConLLU format 
"""


class ConlluData:
    """
    A class to represent a token or a row in CoNLLU format defined in 
    here: https://universaldependencies.org/format.html
    """

    def __init__(self, idx: str, form: str = "_",
                 lemma: str = '_', upos: str = '_',
                 xpos: str = '_', feat: str = '_',
                 head_id: str = '_', deprel: str = '_'):
        """
        Initialize ConnlluData
        """

        self.__idx = idx  # colomn 1
        self.__lemma = lemma  # colomn 2
        self.__form = form  # colomn 3
        self.__upos = upos  # colomn 4
        self.__xpos = xpos  # colomn 5
        self.__feat = feat  # colomn 6
        self.__head_id = head_id  # colomn 7
        self.deprel = deprel  # colomn 8

    def get_id(self) -> str:
        """id getter"""

        return self.__idx

    def get_lemma(self) -> str:
        """lemma getter"""

        return self.__lemma

    def get_form(self) -> str:
        """form getter"""

        return self.__form

    def get_upos(self) -> str:
        """upos getter"""

        return self.__upos

    def get_xpos(self) -> str:
        """xpos getter"""

        return self.__xpos

    def get_feat(self) -> str:
        """feat getter"""

        return self.__feat

    def get_head_id(self) -> str:
        """head_id getter"""

        return self.__head_id

    def get_deprel(self) -> str:
        """dependency relation(deprel) getter"""

        return self.deprel

    def get_conllu_str(self) -> str:
        """ConnluData string representation"""

        the_string = self.__idx + "\t" + self.__form + "\t" + self.__lemma + "\t" + \
                     self.__upos + "\t" + self.__xpos + "\t" + self.__feat + "\t" + \
                     self.__head_id + "\t" + self.deprel + "\t" + "_" + "\t" + "_"

        return the_string

    def __str__(self) -> str:
        return self.get_conllu_str()

    # -------------------------
    # changing value

    def set_form(self, form: str):
        """form setter"""

        self.__form = form

    def set_lemma(self, lemma: str):
        """lemma setter"""

        self.__lemma = lemma

    def set_upos(self, upos: str):
        """upos setter"""

        self.__upos = upos

    def set_feat(self, feat: str):
        """feat setter"""

        self.__feat = feat

    def set_head_id(self, head_id: str):
        """set self.head_id to head_id"""

        self.__head_id = head_id

    def set_deprel(self, deprel):
        """Dependency relation setter"""

        self.deprel = deprel

    def __eq__(self, other):
        """check if 2 objects of ConlluData have the same attributes"""
        if isinstance(other, ConlluData):
            return (
                    self.__idx == other.get_id()
                    and self.__lemma == other.get_lemma()
                    and self.__form == other.get_form()
                    and self.__upos == other.get_upos()
                    and self.__xpos == other.get_xpos()
                    and self.__feat == other.get_feat()
                    and self.__head_id == other.get_head_id()
                    and self.deprel == other.get_deprel()
            )

        return False



"""
Module for aksara dependency parsing feature
"""

import os
import dependency_parsing.core as dep_parser_core

from typing import List, Literal
from aksara.conllu import ConlluData
from aksara.core import analyze_sentence, get_num_lines, split_sentence, sentences_from_file
from aksara.analyzer import BaseAnalyzer
from .utils.conllu_io import write_conllu


class DependencyParser:
    __all_models = [
        "FR_GSD-ID_CSUI",
        "FR_GSD-ID_GSD",
        "IT_ISDT-ID_CSUI",
        "IT_ISDT-ID_GSD",
        "EN_GUM-ID_CSUI",
        "EN_GUM-ID_GSD",
        "SL_SSJ-ID_CSUI",
        "SL_SSJ-ID_GSD"
    ]

    __all_input_modes = ["f", "s"]

    def __init__(self):
        self.default_analyzer = self.__get_default_analyzer()

    def __get_sentence_list(self, input_str, input_mode, sep_regex) -> List[str]:
        if input_mode == "s":
            return split_sentence(input_str, sep_regex)

        if input_mode == "f":
            return sentences_from_file(input_str, sep_regex)

        raise ValueError(
            f"input_mode must be one of {self.__all_input_modes}, but {input_mode} was given"
        )

    def parse(
            self, input_src: str,
            input_mode: Literal["f", "s"] = "s",
            is_informal: bool = False,
            sep_regex: str = None,
            model: str = "FR_GSD-ID_CSUI"
    ) -> List[List[ConlluData]]:

        sentence_list = self.__get_sentence_list(input_src.strip(), input_mode, sep_regex)

        if len(sentence_list) == 1 and sentence_list[0] == "":
            return []

        result = []

        for sentence in sentence_list:
            sentence_result = self._parse_one_sentence(
                sentence, is_informal, model
            )
            result.append(sentence_result)

        return result

    def parse_to_file(
            self, input_src: str,
            write_path: str,
            input_mode: Literal["f", "s"] = "s",
            write_mode: Literal['a', 'w', 'x'] = 'x',
            is_informal: bool = False,
            sep_regex: str = None,
            sep_column: str = '\t',
            model: str = "FR_GSD-ID_CSUI"
    ) -> str:
        result = self.parse(
            input_src, input_mode=input_mode, is_informal=is_informal,
            sep_regex=sep_regex, model=model
        )

        return write_conllu(result, write_path,
                            write_mode=write_mode, separator=sep_column)

    def _parse_one_sentence(
            self, sentence: str,
            is_informal: bool = False,
            model: str = "FR_GSD-ID_CSUI"
    ):
        sentence = sentence.strip()

        if sentence == "":
            return []

        default_dependency_parser = self.__get_default_dependency_parser(model)

        analyzed_sentence = analyze_sentence(
            sentence,
            self.default_analyzer,
            default_dependency_parser,
            v1=False,
            lemma=False,
            postag=False,
            informal=is_informal
        )

        result = []
        for row in analyzed_sentence.split("\n"):
            idx, form, lemma, upos, xpos, feat, head_id, deprel, _, _ = row.split("\t")
            conllu_data = ConlluData(
                idx, form, lemma, upos, xpos, feat, head_id, deprel
            )
            result.append(conllu_data)

        return result

    def __get_default_analyzer(self) -> BaseAnalyzer:
        bin_path = os.path.join(os.path.dirname(__file__), "bin", "aksara@v1.2.0.bin")
        return BaseAnalyzer(bin_path)

    def __get_default_dependency_parser(self, model) -> dep_parser_core.DependencyParser:
        if model not in self.__all_models:
            raise ValueError(f"model must be one of {self.__all_models}, but {model} was given")

        return dep_parser_core.DependencyParser(model)



import unittest

from aksara.conllu import ConlluData


class ConlluTest(unittest.TestCase):

    def setUp(self) -> None:
        self.connlu_data = ConlluData(idx='1', form="berjalan", lemma="jalan",
                                      upos="VERB", xpos='_', feat='Voice=Act',
                                      head_id='0', deprel='root')

        return super().setUp()

    def test_constructor(self):
        self.assertEqual(self.connlu_data.get_id(), '1')
        self.assertEqual(self.connlu_data.get_form(), "berjalan")
        self.assertEqual(self.connlu_data.get_lemma(), "jalan")
        self.assertEqual(self.connlu_data.get_upos(), "VERB")
        self.assertEqual(self.connlu_data.get_xpos(), "_")
        self.assertEqual(self.connlu_data.get_feat(), "Voice=Act")
        self.assertEqual(self.connlu_data.get_head_id(), '0')
        self.assertEqual(self.connlu_data.get_deprel(), 'root')

    def test_setter_getter(self):
        new_form = 'jalan'
        new_lemma = 'jalan'
        new_upos = 'NOUN'
        new_feat = 'Number=Sing'
        new_head_id = '0'
        new_deprel = 'root'

        self.connlu_data.set_form(new_form)
        self.connlu_data.set_lemma(new_lemma)
        self.connlu_data.set_upos(new_upos)
        self.connlu_data.set_feat(new_feat)
        self.connlu_data.set_head_id(new_head_id)
        self.connlu_data.set_deprel(new_deprel)

        self.assertEqual(self.connlu_data.get_id(), '1')
        self.assertEqual(self.connlu_data.get_form(), new_form)
        self.assertEqual(self.connlu_data.get_lemma(), new_lemma)
        self.assertEqual(self.connlu_data.get_upos(), new_upos)
        self.assertEqual(self.connlu_data.get_xpos(), "_")
        self.assertEqual(self.connlu_data.get_feat(), new_feat)
        self.assertEqual(self.connlu_data.get_head_id(), new_head_id)
        self.assertEqual(self.connlu_data.get_deprel(), new_deprel)

    def test_default_underscore_value(self):
        default_val_connlu = ConlluData('1', head_id='0')

        self.assertEqual(default_val_connlu.get_id(), '1')
        self.assertEqual(default_val_connlu.get_form(), '_')
        self.assertEqual(default_val_connlu.get_lemma(), '_')
        self.assertEqual(default_val_connlu.get_upos(), '_')
        self.assertEqual(default_val_connlu.get_xpos(), '_')
        self.assertEqual(default_val_connlu.get_feat(), '_')
        self.assertEqual(default_val_connlu.get_head_id(), '0')
        self.assertEqual(default_val_connlu.get_deprel(), '_')

    def test_get_conllu_str_and__str__must_be_the_same(self):
        self.assertEqual(str(self.connlu_data), self.connlu_data.get_conllu_str())

    def test_get_connllu_str_should_replace_missing_column_with_underscore(self):
        all_default = ConlluData(idx='1', head_id='0')

        num_of_underscore = all_default.get_conllu_str().count('_')
        max_conllu_underscore = 6  # id and headID must not be null
        self.assertGreaterEqual(num_of_underscore, max_conllu_underscore)

    def test_two_conllu_is_equal_if_have_same_attributes(self):
        other_conllu = ConlluData(idx='1', form="berjalan", lemma="jalan",
                                  upos="VERB", xpos='_', feat='Voice=Act',
                                  head_id='0', deprel='root')
        self.assertEqual(self.connlu_data, other_conllu)

    def test_two_conllu_is_not_equal_if_do_not_have_same_attributes(self):
        other_conllu = ConlluData(idx='1', form="jalan", lemma="jalan",
                                  upos="VERB", xpos='_', feat='_',
                                  head_id='0', deprel='root')
        self.assertNotEqual(self.connlu_data, other_conllu)

    def test_conllu_is_not_equal_with_non_conllu(self):
        self.assertNotEqual(self.connlu_data, "")

    def test_multiword_token(self):
        multi_word_conllu = ConlluData(idx='1-2', head_id='0')

        self.assertEqual(multi_word_conllu.get_id(), '1-2')


__author__ = 'max'

import math
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter


class BiLinear(nn.Module):
    '''
    Bi-linear layer
    '''
    def __init__(self, left_features, right_features, out_features, bias=True):
        '''

        Args:
            left_features: size of left input
            right_features: size of right input
            out_features: size of output
            bias: If set to False, the layer will not learn an additive bias.
                Default: True
        '''
        super(BiLinear, self).__init__()
        self.left_features = left_features
        self.right_features = right_features
        self.out_features = out_features

        self.U = Parameter(torch.Tensor(self.out_features, self.left_features, self.right_features))
        self.W_l = Parameter(torch.Tensor(self.out_features, self.left_features))
        self.W_r = Parameter(torch.Tensor(self.out_features, self.left_features))

        if bias:
            self.bias = Parameter(torch.Tensor(out_features))
        else:
            self.register_parameter('bias', None)

        self.reset_parameters()

    def reset_parameters(self):
        nn.init.xavier_uniform_(self.W_l)
        nn.init.xavier_uniform_(self.W_r)
        nn.init.constant_(self.bias, 0.)
        nn.init.xavier_uniform_(self.U)

    def forward(self, input_left, input_right):
        '''

        Args:
            input_left: Tensor
                the left input tensor with shape = [batch1, batch2, ..., left_features]
            input_right: Tensor
                the right input tensor with shape = [batch1, batch2, ..., right_features]

        Returns:

        '''

        left_size = input_left.size()
        right_size = input_right.size()
        assert left_size[:-1] == right_size[:-1], \
            "batch size of left and right inputs mis-match: (%s, %s)" % (left_size[:-1], right_size[:-1])
        batch = int(np.prod(left_size[:-1]))

        # convert left and right input to matrices [batch, left_features], [batch, right_features]
        input_left = input_left.view(batch, self.left_features)
        input_right = input_right.view(batch, self.right_features)

        # output [batch, out_features]
        output = F.bilinear(input_left, input_right, self.U, self.bias)
        output = output + F.linear(input_left, self.W_l, None) + F.linear(input_right, self.W_r, None)
        # convert back to [batch1, batch2, ..., out_features]
        return output.view(left_size[:-1] + (self.out_features, ))

    def __repr__(self):
        return self.__class__.__name__ + ' (' \
               + 'in1_features=' + str(self.left_features) \
               + ', in2_features=' + str(self.right_features) \
               + ', out_features=' + str(self.out_features) + ')'



#!/usr/bin/python3
from tempfile import NamedTemporaryFile
from typing import List
from tqdm import tqdm

import argparse
import mmap
import os
import re
import subprocess
import sys
import codecs

from dependency_parsing import *

from .tokenizer import (
    BaseTokenizer,
)

from .formatter import (
    to_conllu_line,
    to_conllu_line_with_range,
)

from .analyzer import (
    BaseAnalyzer,
)

from .parser import (
    parse,
)

from .disambiguator import Disambiguator

from dependency_parsing.core import DependencyParser

HEADER = """# sent_id = {}
# text = {}
"""

HELP_MSG = {
    "string": "text string",
    "file": "input file",
    "output": "output file",
    "lemma": "only output lemmatization result",
    "postag": "only output POS tagging result",
    "informal": "to use informal rule beside the formal rule",
    "model": "set dependency parser model (default: FR_GSD-ID_CSUI)",
}

base_tokenizer = BaseTokenizer()
disambiguator = Disambiguator()


def analyze_sentence(text, analyzer, dependency_parser, **kwargs):
    surface, SANflags = base_tokenizer.tokenize(text)
    tokens = surface[:]
    flag = dict()

    for key, value in kwargs.items():
        flag[key] = value

    # lowercase first word
    word_pattern = re.compile(r"[^\w\s+]")
    first_word_idx = 0
    while first_word_idx < len(tokens) and word_pattern.match(tokens[first_word_idx]):
        first_word_idx += 1

    # Analyze lemma
    lemma = []
    for i, token in enumerate(tokens):
        temp = token
        if i == first_word_idx:
            temp = token.lower()

        if flag["informal"]:
            temp = "@informal" + temp

        analysis = analyzer.analyze(temp)
        if i == first_word_idx and re.match(r"([A-Za-z]+)(\+X)", analysis):
            analysis = analyzer.analyze(token)

        lemma.append(analysis)

    rows = []
    line_id = 1

    for i in range(len(tokens)):
        temp_lemma = lemma[i].split("\\n")
        temp_lemma = filter(lambda x: x != "", temp_lemma)
        temp_lemma = [temp.split("_") for temp in temp_lemma]
        tmp = []

        # Filter different length lemmas, please handle this case in the future
        min_length = min([len(e) for e in temp_lemma])
        temp_lemma = list(filter(lambda x: len(x) == min_length, temp_lemma))

        for j in range(len(temp_lemma[0])):
            merged = [temp_lemma[k][j] for k in range(len(temp_lemma))]
            tmp.append("\\n".join(merged))
        temp_lemma = tmp

        # temp_lemma = lemma[i].split('_')
        temp_surface = surface[i]
        unsuffixed_pattern = ["PRON", "DET"]

        if len(temp_lemma) == 2:
            is_in_front = any(
                [pattern in temp_lemma[0] for pattern in unsuffixed_pattern]
            )
            split_point = 0

            if is_in_front:
                split_point = len(temp_lemma[0].split("+")[0])
            else:
                split_point = -len(temp_lemma[1].split("+")[0])

            temp_surface = [temp_surface[:split_point], temp_surface[split_point:]]
        else:
            temp_surface = [temp_surface]

        # Add full word line, if splitted
        n_tokens = len(temp_surface)
        if n_tokens > 1:
            new_row = to_conllu_line_with_range(line_id, surface[i], n_tokens)
            rows.append(new_row)

        # Add word line(s)
        for j in range(n_tokens):
            new_row = ""
            if j == n_tokens - 1:
                new_row = to_conllu_line(
                    line_id, temp_surface[j], temp_lemma[j], space_after=SANflags[i]
                )
            else:
                new_row = to_conllu_line(line_id, temp_surface[j], temp_lemma[j])
            rows.append(new_row)
            line_id += 1

    parsed_rows = parse(rows)
    if flag["v1"]:
        parsed_rows = dependency_parser.parse_rows(parsed_rows)
        if flag["lemma"] or flag["postag"]:
            return "\n".join(
                get_lemma_or_postag(parsed_rows, flag["lemma"], flag["postag"])
            )
        else:
            return "\n".join(rows)
    else:
        disambiguated_rows = disambiguator.disambiguate(parsed_rows)
        disambiguated_rows = dependency_parser.parse_rows(disambiguated_rows)
        if flag["lemma"] or flag["postag"]:
            return "\n".join(
                get_lemma_or_postag(disambiguated_rows, flag["lemma"], flag["postag"])
            )
        else:
            return "\n".join(["\t".join(row) for row in disambiguated_rows])


def create_args_parser(bin_file):
    parser = argparse.ArgumentParser(description="Aksara")

    # Add a required, positional argument for the input data file name,
    # and open in 'read' mode
    input_group = parser.add_mutually_exclusive_group(required=True)
    input_group.add_argument("-s", "--string", type=str, help=HELP_MSG["string"])
    input_group.add_argument(
        "-f", "--file", type=argparse.FileType("r"), help="input file"
    )

    # Add optional arguments
    # open in 'write' mode and and specify encoding
    parser.add_argument(
        "--output", type=argparse.FileType("w", encoding="UTF-8"), help=""
    )
    parser.add_argument("--v1", action="store_true")
    parser.add_argument("--lemma", action="store_true", help=HELP_MSG["lemma"])
    parser.add_argument("--postag", action="store_true", help=HELP_MSG["postag"])
    parser.add_argument("--informal", action="store_true", help=HELP_MSG["informal"])
    parser.add_argument("--model", type=str, help=HELP_MSG["model"])

    args = parser.parse_args()
    analyzer = BaseAnalyzer(bin_file)
    if args.model:
        dependency_parser = DependencyParser(args.model)
    else:
        dependency_parser = DependencyParser()

    output = ""
    if args.file:
        with args.file as infile:
            print("Processing inputs...")
            tqdm_setup = tqdm(
                infile,
                total=get_num_lines(infile.name),
                bar_format="{l_bar}{bar:50}{r_bar}{bar:-10b}",
            )
            idx_sentence = 1
            for i, line in enumerate(tqdm_setup, 1):
                text = line.rstrip()
                temp = re.split(r"([.!?]+[\s])", text)
                sentences = []
                for i in range(len(temp)):
                    if i % 2 == 0:
                        sentences.append(
                            temp[i] + (temp[i + 1] if i != len(temp) - 1 else "")
                        )

                for j in range(len(sentences)):
                    temp = analyze_sentence(
                        sentences[j],
                        analyzer,
                        dependency_parser,
                        v1=args.v1,
                        lemma=args.lemma,
                        postag=args.postag,
                        informal=args.informal,
                    )
                    output += HEADER.format(str(idx_sentence), sentences[j], "")
                    output += temp + "\n\n"
                    idx_sentence += 1
    else:
        text = args.string
        temp = re.split(r"([.!?]+[\s])", text)
        sentences = []
        for i in range(len(temp)):
            if i % 2 == 0:
                sentences.append(temp[i] + (temp[i + 1] if i != len(temp) - 1 else ""))

        for i in range(len(sentences)):
            output += HEADER.format(str(i + 1), sentences[i], "")
            output += analyze_sentence(
                sentences[i],
                analyzer,
                dependency_parser,
                v1=args.v1,
                lemma=args.lemma,
                postag=args.postag,
                informal=args.informal,
            )
            output += "\n\n"
        # output += '\n'.join(output)

    output = output.rstrip()
    if args.output:
        args.output.writelines(output)
    else:
        print(output)


def get_num_lines(file_path):
    fp = open(file_path, "r+", encoding="utf-8")
    buf = mmap.mmap(fp.fileno(), 0)
    lines = 0
    while buf.readline():
        lines += 1
    return lines


def get_lemma_or_postag(rows, lemma, postag):
    new_rows = []
    for row in rows:
        temp = [row[0], row[1]]
        if lemma:
            temp.append(row[2])
        if postag:
            temp.append(row[3])
        new_rows.append(temp)

    return ["\t".join(row) for row in new_rows]


def split_sentence(text: str, sep_regex: str = None) -> List[str]:
    """
    this method will split a multi sentences text based on separator regex (sep_regex)

    parameters
    ---------

    text: str
        sentences that will be splitted

    sep_regex: str
        regex rule that determine the points where the text will be splitted at.

    return
    ------
    ReturnType: List of string
        splitted text
    """

    if sep_regex is None:
        sep_regex = r"([.!?]+[\\s])"

    splitted_sentences = re.split(codecs.decode(sep_regex, "unicode_escape"), text)
    sentence_list = []
    for i, sentence in enumerate(splitted_sentences):
        if i % 2 == 0:
            sentence_with_end_mark = sentence + (
                splitted_sentences[i + 1] if i != len(splitted_sentences) - 1 else ""
            )

            sentence_list.append(sentence_with_end_mark)

    return sentence_list


def sentences_from_file(file_path: str, sep_regex: str = None) -> List[str]:
    result = []

    with open(file_path, "r", encoding="utf-8") as infile:
        tqdm_setup = tqdm(
            infile,
            total=get_num_lines(infile.name),
            bar_format="{l_bar}{bar:50}{r_bar}{bar:-10b}",
        )

        for _, line in enumerate(tqdm_setup, 1):
            sentences = split_sentence(line.rstrip(), sep_regex)
            for sentence in sentences:
                result.append(sentence.strip())

    return result



_author__ = 'max'

import logging
import sys


def get_logger(name, filepath=None, level=logging.INFO, handler=sys.stdout,
               formatter='%(asctime)s - %(name)s - %(levelname)s - %(message)s'):
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    formatter = logging.Formatter(formatter)
    stream_handler = logging.StreamHandler(handler)
    stream_handler.setLevel(level)
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)

    if filepath is not None:
        file_handler = logging.FileHandler(filepath, "w")
        file_handler.setLevel(logging.INFO)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    return logger



__author__ = 'max'

import re
MAX_CHAR_LENGTH = 45
NUM_CHAR_PAD = 2

# Regular expressions used to normalize digits.
DIGIT_RE = re.compile(r"\d")

# discard secondary labels
def get_main_deplabel(label):
    return label.split(":")[0]



__author__ = 'max'

from torch.autograd import Variable


def assign_tensor(tensor, val):
    """
    copy val to tensor
    Args:
        tensor: an n-dimensional torch.Tensor or autograd.Variable
        val: an n-dimensional torch.Tensor to fill the tensor with

    Returns:

    """
    if isinstance(tensor, Variable):
        assign_tensor(tensor.data, val)
        return tensor

    return tensor.copy_(val)
